# exlap_utf.rb version 1.20 (encoding: utf-8)  2018/11/13
  # an ruby library for Excel OLE control
  # Copyright (C) T. Yoshiizumi, 2009-2018 All rights reserved.
require "win32ole"
require "csv"
require "nkf"
require "time"

if RUBY_VERSION >= "2.4"
    $exlInt = [Integer]
else
    $exlInt = [Fixnum,Bignum]
end

module ExlUtf
  def to_u(x)
    xc = x.class
    if xc == String
      return NKF.nkf("-m0xw", x)
    elsif xc != Array
      return x
    end
    for i in 0...x.size
      x[i] = to_u(x[i])
    end
    return x
  end

  def to_org(x)
    xc = x.class
    if xc == String
      x = NKF.nkf("-m0xs", x)
      x.force_encoding('Windows-31J')  if RUBY_VERSION >= "1.9"
      return x
    elsif xc != Array
      return x
    end
    for i in 0...x.size
      x[i] = to_org(x[i])
    end
    return x
  end
end

# ワークシート内のセルとその集合を使いやすくするための拡張モジュール
module Exlsheet
  include ExlUtf

  def [] y,x
    tcell = self.Cells.Item(y,x)
    if tcell.MergeCells
      to_u(tcell.MergeArea.Item(1,1).Value)
    else
      to_u(tcell.Value)
    end
  end

  def []= y,x,value
    tcell = self.Cells.Item(y,x)
    if tcell.MergeCells
      tcell.MergeArea.Item(1,1).Value = value
    else
      tcell.Value = value
    end
  end

  def cell(y, x)
    return self.Cells.Item(y,x)
  end

  def to_a()
    if self.empty?
      return []
    end
    y2,x2 = self.range_last
    res = ([y2,x2] == [1,1]) ? [[to_u(self.cell(1,1).Value)]] :
        to_u(self.rip_range(1,1, y2,x2).Value)
    res.map! {|row|
      row.pop  while row.size>0 and row.last==nil
      row
    }
    return res
  end

  def range_to_a(rng = nil)
    if self.empty?
      return []
    end
    rng = self.UsedRange  unless rng
    if (addrs = Exl::yx(rng)) == nil
      return []
    end
    addrs *= 2  if addrs.size == 2
    y1,x1, y2,x2 = addrs
    res = ([y1,x1] == [y2,x2]) ? [[to_u(self.cell(y1,x1).Value)]] : to_u(rng.Value)
    res.map! {|row|
      row.pop  while row.size>0 and row.last==nil
      row
    }
    return res
  end

  def range_address(rng = :NONE, opt = :ABSOLUTE)
    if rng == :NONE
      if self.empty?
        return []
      end
      rng = self.UsedRange
    elsif ($exlInt + [Regexp,String,Float,NilClass]).include?(rng.class)
      rng = self.rip_range(rng)
    end
    res = Exl::yx(rng, opt)
    if !res
      res = []
    end
    return res
  end

  def range_last(rng = :NONE, opt = :ABSOLUTE)  # 所与の領域の末尾番地
    res = self.range_address(rng, opt)
    while res.size > 2
      res.shift
    end
    return res
  end

  def range_start(rng = :NONE, opt = :ABSOLUTE)  # 所与の領域の先頭セル番地
    res = self.range_address(rng, opt)
    while res.size > 2
      res.pop
    end
    return res
  end

  def rip_range(*opt)
    if opt.size == 0
      return self.UsedRange
    end
    types = Array.new
    opt.each {|x|
      if $exlInt.include?(x.class)
        types << :NUMBER
      elsif [String,Regexp].include?(x.class)
        types << :STRING
      else
        types << :OTHER
      end
    }
    tcells = Array.new
    i = 0
    while i < types.size
      if types[i] == :NUMBER and types[i+1] == :NUMBER
        tcells << self.cell(opt[i], opt[i+1])
        i = i + 2
      elsif types[i] == :STRING
        tcells << self.findf(opt[i])
        i = i + 1
      else
        tcells.clear
        break
      end
    end
    rng = nil
    if tcells.size == 1
      y,x = self.range_last
      tcells << self.cell(y,x)
    end
    if tcells.size == 2 and tcells[0] == tcells[1]
      rng = tcells[0]
    elsif tcells.size == 2
      rng = self.Range(tcells[0], tcells[1])
    end
    return rng
  end
  alias :rr :rip_range

  def range_pole(rng = :NONE)  # 両極のセルオブジェクトを返す
    res = []
    w = self.range_address(rng)
    if w.size == 2
      res << self.Cells.Item(w[0], w[1])
    elsif w.size == 4
      res << self.Cells.Item(w[0], w[1])
      res << self.Cells.Item(w[2], w[3])
    end
    return res
  end
  alias :rp :range_pole
  alias :range_pair :range_pole

  def entire_range  # 使われている領域全体(始点A1)を返す
    y,x = self.range_last
    self.Range(self.Cells.Item(1,1), self.Cells.Item(y,x))
  end

  def range_autofit(rng = nil)
    rng = self.UsedRange  unless rng
    rng.Columns.each {|col|  col.AutoFit}
    rng.Rows.each {|row|  row.AutoFit}
  end

  def range_allbox(rng=nil, opt={})
    unless rng
      if self.empty?
        return
      end
      rng = self.UsedRange
    end
    opt['LineStyle'] = 1  unless opt['LineStyle']  # XlContinuous 実線
    opt['Weight'] = 2  unless opt['Weight']  # XlThin 細線
    opt['ColorIndex'] = -4105 unless opt['ColorIndex'] # XlAutomatic 自動設定
    bd = rng.Borders
    bd.LineStyle = opt['LineStyle']
    bd.Weight = opt['Weight']
    bd.ColorIndex = opt['ColorIndex']
  end

  def range_doublebox(rng=nil, opt={})
    unless rng
      if self.empty?
        return
      end
      rng = self.UsedRange
    end
    opt['LineStyle'] = 1  unless opt['LineStyle']  # XlContinuous 実線
    opt['Weight'] = 2  unless opt['Weight']  # XlThin 細線
    opt['ColorIndex'] = -4105 unless opt['ColorIndex'] # XlAutomatic 自動設定
    opt['Frame'] = -4138  unless opt['Frame']  # XlMedium 外枠・太線
        # 上段
    bd1 = rng.Borders(XlEdgeTop)
    bd1.LineStyle = opt['LineStyle']
    bd1.Weight = opt['Frame']
    bd1.ColorIndex = opt['ColorIndex']
        # 下段
    bd2 = rng.Borders(XlEdgeBottom)
    bd2.LineStyle = opt['LineStyle']
    bd2.Weight = opt['Frame']
    bd2.ColorIndex = opt['ColorIndex']
        # 左側
    bd3 = rng.Borders(XlEdgeLeft)
    bd3.LineStyle = opt['LineStyle']
    bd3.Weight = opt['Frame']
    bd3.ColorIndex = opt['ColorIndex']
        # 右側
    bd4 = rng.Borders(XlEdgeRight)
    bd4.LineStyle = opt['LineStyle']
    bd4.Weight = opt['Frame']
    bd4.ColorIndex = opt['ColorIndex']
        # 内側の縦線
    bd5 = rng.Borders(XlInsideVertical)
    bd5.LineStyle = opt['LineStyle']
    bd5.Weight = opt['Weight']  # 細線
    bd5.ColorIndex = opt['ColorIndex']
        # 内側の横線
    bd6 = rng.Borders(XlInsideHorizontal)
    bd6.LineStyle = opt['LineStyle']
    bd6.Weight = opt['Weight']
    bd6.ColorIndex = opt['ColorIndex']
  end

  def range_clearbox(rng=nil)
    rng = self.UsedRange  unless rng
    rng.Borders.LineStyle = -4142  # XlLineStyleNone(罫線を消す)
  end

  def findf(key, prop=nil)  # propはValueなどのプロパティ
    tcell = nil
    if prop == nil or prop.downcase == 'value'
      if key.class == Regexp
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key =~ to_u(col.Value).to_s
              tcell = col
              break
            end
          }
          break  if tcell
        }
      elsif ($exlInt + [String,Float,NilClass,TrueClass,
          FalseClass]).include?(key.class)
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key == to_u(col.Value)
              tcell = col
              break
            end
          }
          break  if tcell
        }
      end
    else
      if key.class == Regexp
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key =~ to_u(eval("col.#{prop}.to_s"))
              tcell = col
              break
            end
          }
          break  if tcell
        }
      elsif ($exlInt + [String,Float,NilClass,TrueClass,
          FalseClass]).include?(key.class)
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key == to_u(eval("col.#{prop}"))
              tcell = col
              break
            end
          }
          break  if tcell
        }
      end
    end
    return tcell
  end

  def finda(key, prop=nil, &block)
    res = []
    if prop == nil or prop.downcase == 'value'
      if key.class == Regexp
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key =~ to_u(col.Value).to_s
              if block
                block.call(col)
              else
                res << col
              end
            end
          }
        }
      elsif ($exlInt + [String,Float,NilClass,TrueClass,
          FalseClass]).include?(key.class)
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key == to_u(col.Value)
              if block
                block.call(col)
              else
                res << col
              end
            end
          }
        }
      end
    else
      if key.class == Regexp
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key =~ to_u(eval("col.#{prop}.to_s"))
              if block
                block.call(col)
              else
                res << col
              end
            end
          }
        }
      elsif ($exlInt + [String,Float,NilClass,TrueClass,
          FalseClass]).include?(key.class)
        self.UsedRange.Rows.each {|row|
          row.Columns.each {|col|
            if key == to_u(eval("col.#{prop}"))
              if block
                block.call(col)
              else
                res << col
              end
            end
          }
        }
      end
    end
    return res
  end

  def hyperlink(*opt)
    tcell = nil
    uri = nil
    text = nil
    if $exlInt.include?(opt[0].class) and [3,4].include?(opt.size)
      begin
        tcell = self.Cells.Item(opt[0], opt[1])
      rescue
        tcell = nil
      end
      uri = opt[2]
      text = opt[3]  if opt.size == 4
    elsif opt[0].class == String and opt[0] =~ /^[A-Z]+\d+$/i and
          [2,3].include?(opt.size)
      begin
        tcell = self.Range(opt[0].upcase)
      rescue
        tcell = nil
      end
      uri = opt[1]
      text = opt[2]  if opt.size == 3
    end
    if tcell == nil or uri == nil
      STDERR.printf("Exlsheet.hyperlink(): invalid parameters (%s)\n\n",
          opt.join(","))
      return
    end
    if text
      hs = {'Anchor'=>tcell, 'Address'=>uri, 'TextToDisplay'=>text}
      self.Hyperlinks.Add(hs)
    else
      hs = {'Anchor'=>tcell, 'Address'=>uri}
      self.Hyperlinks.Add(hs)
    end
  end

  def empty?  # 空のセルだけか？
    res = false
    if Exl::yx(self.UsedRange) == [1,1] and
         to_u(self.Cells.Item(1,1).Value) == nil and
         self.Cells.Item(1,1).Formula.to_s == ''
      res = true
    end
    return res
  end

  def empty_sheet?  # ワークシートが空か？
    res = false
    if self.empty? and
         self.ChartObjects.Count == 0 and self.Shapes.Count == 0 and
         self.Hyperlinks.Count == 0 and self.QueryTables.Count == 0 and
         self.Scenarios.Count == 0
      res = true
    end
    return res
  end

  # Range内の各セルを一つづつ辿る。セルオブジェクトを返す。
  def range_each(rng = nil, type = :CELL, &block)
    unless block
      return
    end
    rng = :NONE  unless rng
    ary = self.range_address(rng)
    if ary.size == 0
      return
    elsif ary.size == 2
      ary *= 2
    end
    y1,x1, y2,x2 = ary
    for y in y1..y2 do
      for x in x1..x2 do
        if type == :CELL
          tcell = self.cell(y,x)
          block.call(tcell)
        else
          block.call(y, x)
        end
      end
    end
  end

  def from_a(*arg, &block)
    fmt_sw = nil
    if arg.include?(:f) or arg.include?(:format)
      fmt_sw = true
      arg.delete(:f)
      arg.delete(:format)
    end
    y1 = x1 = ary = nil
    if arg.size == 2 and arg[0].class == String and arg[1].class == Array
      y1,x1 = Exl::yx(self.Range(arg[0]))
      ary = arg[1]
    elsif arg.size == 3 and arg[0].class == Fixnum and
          arg[1].class == Fixnum and arg[2].class == Array
      y1 = arg[0]
      x1 = arg[1]
      ary = arg[2]
    elsif arg.size == 1 and arg[0].class == Array
      y1 = x1 = 1
      ary = arg[0]
    end
    if !y1 or !x1 or !ary or ary.size == 0
      return nil
    end
    ary = [ary]  if ary[0].class != Array
    max_x = 0
    ary.each {|row|  m=row.size; max_x=m if m>max_x}
    if max_x == 0
      return nil
    end
    ary2 = []
    ary.each {|row|  ary2.push(row + [nil]*(max_x-row.size))}
    max_y = ary2.size
    rng = rip_range(y1,x1, y1+max_y-1,x1+max_x-1)
    rng.Value = ary2
    if fmt_sw
      ary.each_with_index do |row, y|
        ary2[y] = []
        row.each_with_index do |val, x|
          fmt = nil  # NumberFormatLocal の設定文字列をこれにセット
          ii = ff = nil  # ii:整数部, ff:小数部
          val = val.to_s
          if val =~ /^-?(\d[0-9,]*)\.(\d+)$/  # 小数点数
            ii = $1
            ff = $2
          elsif val =~ /^-?(\d[0-9,]*)$/  # 整数
            ii = $1
          end
          if ii
            if ii =~ /^0/
              fmt = ii.gsub(/\d/, "0")
            else
              fmt = ii.gsub(/\d/, "#")
              fmt = fmt.sub(/\#$/, "0")
              fmt = fmt.sub(/^#+/, "")  unless fmt =~ /,/
            end
            if fmt =~ /^[0#,]+([0#],[0#]{3})$/
              fmt = $1
            end
            if ff
              fmt = fmt + '.' + ff.gsub(/\d/, "0")
            end
            self.cell(y1+y, x1+x).NumberFormatLocal = fmt
            if ii and !ff
              ary2[y][x] = 0
            elsif ii and ff
              ary2[y][x] = ff.length
            end
          end
        end
      end
      for x in 0...max_x
        max_ff = 0
        for y in 0...max_y
          if ary2[y][x].class == Fixnum and ary2[y][x] > max_ff
            max_ff = ary2[y][x]
          end
        end
        if max_ff > 1
          for y in 0...max_y
            next  if !ary2[y][x] or ary2[y][x] == 0 or ary2[y][x] == max_ff
            fmt = self.cell(y1+y, x1+x).NumberFormatLocal
            fmt = fmt + "?" * (max_ff - ary2[y][x])
            self.cell(y1+y, x1+x).NumberFormatLocal = fmt
          end
        end
      end
    end
    if block
      ary.each_with_index {|row, y|
        row.each_with_index {|val, x|
          block.call(y1+y, x1+x, val)
        }
      }
    end
    return rng
  end

  def from_f(*arg, &block)
    asz = arg.size
    fname = nil
    while asz > 0
      if arg[asz-1].class == String
        fname = arg[asz-1]
        break
      end
      asz -= 1
    end
    if fname.class != String or !test(?f, fname) or test(?z, fname)
      STDERR.puts "Exlsheet.from_f: can't open or read '#{fname}'."
      return nil
    end
    str = Exl::file_read(fname)
    if str.to_s == ''
      return nil
    end
    ary = Exl::str2ary(str, nil, nil)
    ary.map! {|row|
      row.map! {|x|
        x = x.to_s
        if x =~ /^\s*$/
          x = nil
        end
        x
      }
    }
    arg[asz-1] = ary
    self.from_a(*arg, &block)
  end

  def save(filename, type=nil)
    if type == nil
      case File.extname(filename).downcase
      when '.csv'  then  type = :csv;
      when '.txt'  then  type = :txt;
      when '.prn'  then  type = :prn;
      end
    end
    if [:csv, :txt, :tab, :prn].include?(type)
      fname = Exl::fullpath(filename)
      preval = self.Application.DisplayAlerts
      self.Application.DisplayAlerts = false
      case type
      when :csv, :prn
        self.SaveAs({'Filename'=>"#{fname}",
            'FileFormat'=>((type == :csv) ? XlCSV : XlTextPrinter),
            'CreateBackup'=>false})
      when :txt, :tab
        ws = self.Parent.ActiveSheet
        self.Activate
        self.Parent.SaveAs({'Filename'=>"#{fname}",
            'FileFormat'=>XlText, 'CreateBackup'=>false})
        ws.Activate
      end
      self.Application.DisplayAlerts = preval
    end
  end
end


# Excel関係の定数を収納するためのモジュール(空っぽでかまわない)
module ExlConstant
end

def xl_load()
  begin
    xlw = WIN32OLE.new('Excel.Application')
  rescue
    xlw = nil
  end
  if xlw
    xlw.Visible = false
    unless defined?(ExlConstant::CONSTANTS)
      WIN32OLE.const_load(xlw, ExlConstant)
    end
    xlw.Quit
  end
end

# Excelワークブックを扱うためのクラス
class Exlwb
  include ExlConstant
  include ExlUtf

    TypeExp = {1=>'Standard', 2=>'Class', 3=>'UserForm'}
    Fext = {1=>'.bas', 2=>'.cls', 3=>'.frm'}

  def initialize(xl)
    @app = xl
    @obj = @open_type = @filename = nil
  end
  attr_accessor :app
  attr_accessor :obj
  attr_accessor :open_type
  attr_accessor :filename

  def close()
    @app.book_close(self)
  end

  # ファイル名の拡張子が .xls などになるよう調整
  def filename_arrange(fname)
    w = fname.gsub(/\\/, "/")
    e = File.extname(w)
    d = File.dirname(w)
    f = File.basename(w, ".*")
    if d != ""
      d = d.sub(/\/$/, "")
      d = d.gsub(/\//, "\\") + "\\"
    end
    res = d + f
    if @app.ext_type[e.downcase] != nil
      res = res + e
    else
      res = res + ".xls"
    end
    return res
  end

  def save(save_name = nil)
    return  unless @obj
    ext_type = @app.ext_type
    da_v = @app.display_alert(false)
    error_msg = nil
    if save_name == nil
      if @open_type == 'OPEN'
        ffmt = ext_type[File.extname(@filename).downcase]
        if ffmt
          begin
            @obj.SaveAs({'FileName'=>@filename, 'FileFormat'=>ffmt})
          rescue
            error_msg = $!
          end
        else
          begin
            @obj.Save  # 上書き保存
          rescue
            error_msg = $!
          end
        end
      else
        outfile = filename_arrange(@filename)
        ffmt = ext_type[File.extname(outfile).downcase]
        begin
          @obj.SaveAs({'FileName'=>outfile, 'FileFormat'=>ffmt})
        rescue
          error_msg = $!
        end
      end
    else  # save_nameが指定された場合
      outfile = Exl::getAbsolutePath(save_name)
      ffmt = ext_type[File.extname(outfile).downcase]
      unless ffmt
        outfile = filename_arrange(outfile)
        ffmt = ext_type[File.extname(outfile).downcase]
      end
      begin
        @obj.SaveAs({'FileName'=>outfile, 'FileFormat'=>ffmt})
      rescue
        error_msg = $!
      end
    end
    @app.display_alert(da_v)
    if error_msg
      svname = !save_name ? self.name : save_name
      STDERR.printf("Exlwb.save: can't save '%s'\n%s\n\n", svname, error_msg)
    end
  end

  def name  # ワークブックの名前(フルパスでない・拡張子を含む)を返す
    return  File.basename(@filename)
  end

  def sheet_names()
    res = []
    return res  unless @obj
    for i in 1..@obj.Worksheets.Count
      res << to_u(@obj.Worksheets.Item(i).Name)
    end
    return res
  end

  def sheet_number(sheet)
    sheet = select_sheet(sheet)  if sheet.class == String
    i = self.sheet_names.index(to_u(sheet.Name))
    if i == nil
      i = 0
    else
      i += 1
    end
    return i
  end

  def offset_sheet(sheet, diff)
    rsheet = nil
    n = self.sheet_number(sheet)
    if n > 0 and (n + diff) > 0
      rsheet = self.select_sheet(n + diff)
    end
    return rsheet
  end

  def select_sheet(target)
    sheet = nil
    return sheet  unless @obj
    begin
      sheet = @obj.Worksheets.Item(target)
      sheet.extend Exlsheet
    rescue
      sheet = nil
    end
    return sheet
  end
  alias  :ss  :select_sheet

  def last_sheet()
    target = self.sheet_names.size
    return select_sheet(target)
  end

  def add_sheet()
    begin
      sobj = @obj.Worksheets.Add({'After'=>last_sheet()})
      sobj.extend Exlsheet
    rescue
      sobj = nil
    end
    return sobj
  end

  def insert_sheet(target)
    target = select_sheet(target)  if target.class == String
    before = sheet_names()
    begin
      @obj.Worksheets.Add({'Before'=>target})
      fore = sheet_names()
      w = (fore - before)
      sn = w.shift  if w != nil and w.size > 0
    rescue
      sn = nil
    end
    return (sn != nil) ? select_sheet(sn) : nil
  end

  def delete_sheet(target)
    if @obj.Worksheets.Count > 1
      @app.display_alert(false) do
        sheet = (target.class == String) ? select_sheet(target) : target
        sheet.Delete
      end
    end
  end

  def empty_sheet_names()
    res = []
    aa = self.sheet_names
    aa.each {|sn|
      s = self.select_sheet(sn)
      if s.empty_sheet?
        res << sn
      end
    }
    return res
  end

  def first_empty_sheet()
    sheet = nil
    ary = self.empty_sheet_names  # 空っぽのシート(複数)の名前を配列で取得
    if ary.size > 0  # 空っぽのシートが存在する場合
      sn = ary.shift  # 最初の空っぽシートの名前を得る
      sheet = self.select_sheet sn  # 名前がsnのシートを処理対象にする
    else  # 空っぽのシートが存在しない場合
      sheet = self.add_sheet  # シートを追加してその名前を得る
    end
    return sheet
  end
  alias :fes :first_empty_sheet

  def each(&block)
    return  unless block
    self.sheet_names.each {|sn|
      block.call(self.select_sheet(sn))
    }
  end

  def macro_code(type=nil, name=nil)  # componentのsource codeを取得
    str = ''
    @obj.VBProject.VBComponents.each do |component|
      next  if ![1,2,3].include?(component.Type) or
          type and component.Type != type or name and to_u(component.Name) != name
      str += "' <<type:#{TypeExp[component.Type]},  name:#{component.Name}>>\n"
      cnt = component.CodeModule.CountOfLines
      cstr = (cnt > 0) ? component.CodeModule.Lines(1, cnt) :
          "' ---- empty ----"
      str += "#{cstr}\n\n"
      break  if type and name
    end
    return to_u(str.gsub(/\r\n/, "\n"))
  end

  def macro_add(str, type=nil, name=nil)  # componentを追加
    str = (str.class != String or str =~ /\A\s*\Z/) ? nil : str.dup
    fname = nil
    if str and str !~ /\r|\n/ and
          test(?e, Exl::getAbsolutePath(str))
      fname = Exl::getAbsolutePath(str)
    end
    type = 1  unless type
    component = @obj.VBProject.VBComponents.Add(type)
    if name
      begin
        component.Name = name
      rescue
        STDERR.printf("Exlwb.macro_add: can't be named as '%s'\n\n", name)
      end
    end
    if str
      str = str.gsub(/\' <<type\:[A-Za-z]+, +name\:.+?>>\n/, "")
      str = str.gsub(/' ---- empty ----\n/, "")
      !fname ? component.CodeModule.AddFromString(str) :
          component.CodeModule.AddFromFile(fname)
    end
    return component
  end

  def macro_export(type=nil, name=nil, opt=nil)  # componentをfileとして出力
    len = self.name.length
    path = @filename[0...-len]
    msg = {}
    res = []
    @obj.VBProject.VBComponents.each do |component|
      next  if ![1,2,3].include?(component.Type) or
          type and component.Type != type or name and to_u(component.Name) != name
      fname = !opt ? "#{path}#{component.Name}#{Fext[component.Type]}" :
          "#{path}#{self.name}!#{component.Name}#{Fext[component.Type]}"
      begin
        component.Export(fname)
      rescue
        msg[fname] = $!
      end
      res << [component.Type, to_u(component.Name)]  if !msg[fname]
    end
    msg.each do |fname, str|
      STDERR.printf("Exlwb.macro_export: fails to export '%s'\n%s\n\n",
          File.basename(fname), str)
    end
    return res
  end

  def macro_import(filename, dup_check=false)  # componentをfileから取得
    res = nil
    filename = Exl::getAbsolutePath(filename)
    unless test(?e, filename)
      return res
    end
    bodyname = File.basename(filename, ".*")
    msg = nil
    exist_flag = false
    if dup_check
      @obj.VBProject.VBComponents.each do |component|
        next  unless [1,2,3].include?(component.Type)
        if bodyname.casecmp(component.Name) == 0
          msg = sprintf("Exlwb.macro_import: '%s' has already existed " +
              "in '%s'.\n\n", to_u(component.Name), to_u(@obj.Name))
          exist_flag = true
          break
        end
      end
    end
    if !dup_check or !exist_flag
      begin
        component = @obj.VBProject.VBComponents.Import(filename)
      rescue
        component = nil
        msg = sprintf("Exlwb.macro_import: error!\n%s\n\n", $!)
      end
      res = [component.Type, to_u(component.Name)]  if component
    end
    if msg
      STDERR.print msg
    end
    return res
  end

  def macro_remove(type=nil, name=nil)  # componentを削除
    res = []
    project = @obj.VBProject
    project.VBComponents.each do |component|
      next  if ![1,2,3].include?(component.Type) or
          type and component.Type != type or name and to_u(component.Name) != name
      tn = [component.Type, to_u(component.Name)]
      begin
        project.VBComponents.Remove(component)
      rescue
        tn = nil
      end
      res << tn  if tn
      break  if type and name
    end
    return res
  end

  def macro_list(type=nil, name=nil)  # componentのType, Nameの組みを配列で返す
    res = []
    @obj.VBProject.VBComponents.each do |component|
      next  if ![1,2,3].include?(component.Type) or
          type and component.Type != type or name and to_u(component.Name) != name
      res << [component.Type, to_u(component.Name)]
      break  if type and name
    end
    return res
  end

  def macro_list2(type=nil, name=nil)  # component objectを配列で返す
    res = []
    @obj.VBProject.VBComponents.each do |component|
      next  if ![1,2,3].include?(component.Type) or
          type and component.Type != type or name and to_u(component.Name) != name
      res << component
      break  if type and name
    end
    return res
  end

  def macro_run(mname, *arg)  # マクロを実行
    mname = to_org(mname)
    arg = to_org(arg)
    cmd = "'#{@obj.Name}'!#{mname}"
    msg = nil
    res = nil
    begin
      if arg.size == 0
        res = @app.obj.Run cmd
      else
        res = @app.obj.Run(cmd, *arg)
      end
    rescue
      msg = sprintf("Exlwb.macro_run: error!\n%s\n\n", $!)
      res = nil
    end
    if msg
      STDERR.print msg
    end
    return res
  end
  alias :run :macro_run

  def macro_name(type=nil, name=nil)  # component内のprocedure name一覧
    res = {}
    @obj.VBProject.VBComponents.each do |component|
      next  if ![1,2,3].include?(component.Type) or
          type and component.Type != type or name and to_u(component.Name) != name
      proc_names = []
      ccm = component.CodeModule
      cnt = ccm.CountOfLines
      for i in 1..cnt
        pn = ccm.ProcOfLine(i,0)
        if pn.class == String  and pn != "" and !proc_names.include?(pn)
          proc_names << pn
        end
      end
      res[[component.Type, to_u(component.Name)]] = proc_names
      break  if type and name
    end
    return res
  end

  def pdf_out(arg = {})
    hs = {'Type'=>XlTypePDF}
    if arg.class == String
      arg = hs.update({'Filename'=>arg})
    elsif arg.class == Hash
      arg = hs.update(arg)
    end
    if arg.class == Hash
      fname = arg['Filename']
      unless fname
        fname = @filename.sub(/\.[^.]*$/, "") + ".pdf"
      end
      arg['Filename'] = Exl::fullpath(fname)
      @obj.ExportAsFixedFormat(arg)
    else
      STDERR.puts "invalid parameter of 'Exlwb::pdf_out':\n" +
          "  #{arg.inspect}"
    end
  end
  alias :pdf :pdf_out

  def method_missing(sym, *arg)
    name = sym.to_s
    if name =~ /^[A-Z_]/
      return @obj.send(sym, *arg)
    else
      STDERR.puts "undefined method `#{name}' for Wrdoc (NameError)"
    end
  end
end


# Excel Applicationを扱うためのクラス
class Exlap
  include ExlConstant
  include ExlUtf

  def init_set()
    @obj = @excel_version = @ext_type = nil
    @order = @last = nil
    @auto_quit = false
    @auto_close = false
    @books = []
    begin
      @obj = WIN32OLE.new('Excel.Application')
    rescue
      @obj = nil
    end
    if @obj
      @obj.Visible = true
      unless defined?(ExlConstant::CONSTANTS)
        WIN32OLE.const_load(@obj, ExlConstant)
      end
      @excel_version = @obj.Version.to_f
      if @excel_version >= 12.0
        @ext_type = {".xls"=>XlWorkbookNormal, ".xlsx"=>XlOpenXMLWorkbook,
                     ".xlsm"=>XlOpenXMLWorkbookMacroEnabled,
                     ".htm"=>XlHtml, ".html"=>XlHtml,
                     ".xml"=>XlXMLSpreadsheet,
                     ".ods"=>XlOpenDocumentSpreadsheet}
      else
        @ext_type = {".xls"=>XlWorkbookNormal,
                     ".xml"=>XlXMLSpreadsheet,
                     ".htm"=>XlHtml, ".html"=>XlHtml}
        if defined?(XlOpenDocumentSpreadsheet)
          @ext_type[".ods"] = XlOpenDocumentSpreadsheet
        end
      end
    end
  end
  private :init_set
  attr_reader :obj
  attr_reader :excel_version
  attr_reader :ext_type
  attr_reader :order
  attr_reader :last
  attr_accessor :auto_quit
  attr_accessor :auto_close

  def Exlap.callback
    proc { $ExlapObject.quit()  if $ExlapObject != nil }
  end

  def initialize(*fname, &block)
    ObjectSpace.define_finalizer(self, Exlap.callback)
    init_set()
    $ExlapObject = self  if @obj
    if @obj and fname.size > 0
      @last = fname.size
      for n in 0...@last
        @order = n + 1
        wb = book_open(fname[n])
        if block
          block.call(wb)
          book_close(wb)  if @auto_close
        end
      end
      @order = @last = nil
      self.quit  if block
    end
  end

  def quit()
    if @obj
      while wb = @books.shift
        book_close(wb)
        wb = nil
      end
      display_alert(false)
      @obj.Workbooks.Close
      @obj.Quit
      @books = []
      @obj = @excel_version = @ext_type = nil
      @order = @last = nil
      $ExlapObject = nil
    end
  end

  def opens(*fname, &block)
    res = []
    if @obj and fname.size > 0
      @last = fname.size
      for n in 0...@last
        @order = n + 1
        wb = book_open(fname[n])
        if block
          block.call(wb)
          book_close(wb)  if @auto_close
        else
          res << wb
        end
      end
      @order = @last = nil
      self.quit  if block and @auto_quit
    end
    if res.size < 1
      res = nil
    end
    return res
  end

  def opens_once(*fname, &block)
    unless block
      return
    end
    res = []
    if @obj and fname.size > 0
      for n in 0...fname.size
        wb = book_open(fname[n])
        res << wb
      end
    end
    if res.size > 0
      block.call(*res)
      if @auto_close
        res.each {|wb|
          book_close(wb)
        }
      end
      self.quit  if @auto_quit
    end
  end

  def book_open(param)
    wb = nil
    fname = nil
    if param
      if param.class == String
        fname = param.dup
        param = Hash.new
      elsif param.class == Hash
        fname = param['Filename']
      else
        param = nil
      end
    end
    if @obj and param and fname
      fname = Exl::getAbsolutePath(fname)
      param['Filename'] = fname
      wb = select_wb(fname)
      unless wb  # まだ開いていないブックである
        wb = Exlwb.new(self)
        if test(?e, fname)  # 既存のファイルをオープン
          display_alert(false) do
            wb.obj = @obj.Workbooks.Open(param)
            wb.open_type = 'OPEN'
          end
        else  # 存在しないファイルなら新規作成
          wb.obj = @obj.Workbooks.Add()
          wb.open_type = 'ADD'
        end
        wb.filename = fname
      end
      if wb
        @books << wb
      end
    end
    return wb
  end

  def book_close(wb)
    if wb
      if wb.obj
        da_v = self.display_alert(false)
        wb.obj.Close
        self.display_alert(da_v)
      end
      wb.app = wb.obj = wb.open_type = wb.filename = nil
      @books.delete(wb)
    end
  end

  def book_opentext(param)
    wb = nil
    fname = nil
    if param
      if param.class == String
        fname = param.dup
        param = Hash.new
      elsif param.class == Hash
        fname = param['Filename']
      else
        param = nil
      end
    end
    if @obj and param and fname
      fname = Exl::getAbsolutePath(fname)
      param['Filename'] = fname
      wb = select_wb(fname)
      unless wb  # まだ開いていないブックである
        res = false
        display_alert(false) do
          if test(?e, fname)
            res = @obj.Workbooks.OpenText(param)
          end
        end
        if res == true
          wb = Exlwb.new(self)
          n = @obj.Workbooks.Count
          wb.obj = @obj.Workbooks(n)
          wb.filename = fname
          wb.open_type = 'OPEN'
        end
      end
      if wb
        @books << wb
      end
    end
    return wb
  end

  def select_wb(fname)  # フルパスのfnameと照合するワークブックを選択
    wb = nil
    @books.each {|b|
      if b.filename.casecmp(fname) == 0
        wb = b
        break
      end
    }
    return wb
  end
  private :select_wb

  def select_book(target)
    wb = nil
    if target.class == String
      @books.each {|b|
        if b.name.casecmp(target) == 0
          wb = b
          break
        end
      }
    elsif $exlInt.include?(target.class)
      count = 0
      @books.each {|b|
        count += 1
        if count == target
          wb = b
          break
        end
      }
    end
    return wb
  end

  def book_names()
    res = []
    @books.each {|b|
      res << b.name
    }
    return res
  end

  def display_alert(val, &block)
    before_val = @obj.DisplayAlerts
    @obj.DisplayAlerts = val
    if block
      block.call()
      @obj.DisplayAlerts = before_val
      return
    end
    return before_val
  end

  def each(&block)
    return  unless block
    @last = @books.size
    @order = 0
    @books.each {|wb|
      @order += 1
      block.call(wb)
    }
    @order = @last = nil
  end

  def version()  # exlapのバージョン
    return "1.18"
  end

  def addin_list(regex = nil)
    ary = []
    @obj.AddIns.each do |target|
      if regex.class != Regexp or regex =~ to_u(target.Name)
        ary << [to_u(target.Name), to_u(target.FullName)]
      end
    end
    return ary
  end

  def method_missing(sym, *arg)
    name = sym.to_s
    if name =~ /^[A-Z_]/
      return @obj.send(sym, *arg)
    else
      STDERR.puts "undefined method `#{name}' for Wrdoc (NameError)"
    end
  end
end

# その他、クラスに属さないメソッド
module Exl
  extend ExlUtf

  module_function()
  def getAbsolutePath(fname)  # 指定のファイル名をフルパスで返す
    fso = WIN32OLE.new('Scripting.FileSystemObject')
    return fso.GetAbsolutePathName(fname)
  end

  def fullpath(fname)
    fname = fname.gsub(/\//, "\\")
    fso = WIN32OLE.new('Scripting.FileSystemObject')
    return fso.GetAbsolutePathName(fname)
  end

  # excelのセルまたはセル群の番地を返す
  def yx(rng, opt = :ABSOLUTE)  # yx(rc)座標の数値で返す
    res = nil
    opt_v = opt == :ABSOLUTE ? true : false
    begin
      a = rng.Address({'RowAbsolute'=>opt_v, 'ColumnAbsolute'=>opt_v,
          'ReferenceStyle'=>-4150})  # -4150 = XlR1C1
    rescue
      a = nil
    end
    if a
      if a =~ /^R(\d+)C(\d+)$/i
        res = [$1.to_i, $2.to_i]
      elsif a =~ /^R(\d+)C(\d+)\:R(\d+)C(\d+)$/i
        res = [$1.to_i, $2.to_i, $3.to_i, $4.to_i]
      elsif a =~ /^R\[(\d+)\]C\[(\d+)\]$/i
        res = [$1.to_i, $2.to_i]
      elsif a =~ /^R\[(\d+)\]C\[(\d+)\]\:R\[(\d+)\]C\[(\d+)\]$/i
        res = [$1.to_i, $2.to_i, $3.to_i, $4.to_i]
      end
    else
      res = nil
    end
    return res
  end

  def rc(rng, opt = :ABSOLUTE)
    return yx(rng, opt)
  end

  def ab(rng, opt=:ABSOLUTE)  # $A$1:$D$4 などの形式で返す
    opt_v = opt == :ABSOLUTE ? true : false
    begin
      res = rng.Address({'RowAbsolute'=>opt_v, 'ColumnAbsolute'=>opt_v,
          'ReferenceStyle'=>1})  # 1=XlA1
    rescue
      res = nil
    end
    return res
  end

  def a1(rng)
    res = ab(rng)
    if res
      res = res.gsub(/\$/, "")
    end
    return res
  end

  # 現在位置curから見た時の rngの相対番地を
    # RC形式(R[2]C[2], R[1]C[1]:R[3]C[3] など)で返す。
  def relrc(rng, cur)
    rng_ary = (rng.class == Array) ? rng : yx(rng)
    cur_ary = (cur.class == Array) ? cur : yx(cur)
    ary = []
    rng_ary.each_with_index do |val,i|
      if (i % 2) == 0  # 行番号:yの扱い
        ary << (val - cur_ary[0])
      else  # 列番号:xの扱い
        ary << (val - cur_ary[1])
      end
    end
    res = ''
    case ary.size
    when 2  # セル一つ
      res = sprintf("R[%d]C[%d]", ary[0], ary[1])
    when 4  # 範囲
      res = sprintf("R[%d]C[%d]:R[%d]C[%d]", ary[0], ary[1], ary[2], ary[3])
    end
    return res.gsub(/\[0\]/, "")
  end

  # 領域rngの絶対番地をRC形式:R2C2, R1C1:R3C3 などで返す
  def absrc(rng)
    ary = (rng.class == Array) ? rng : yx(rng)
    res = ''
    case ary.size
    when 2  # セル一つ
      res = sprintf("R%dC%d", ary[0], ary[1])
    when 4  # 範囲
      res = sprintf("R%dC%d:R%dC%d", ary[0], ary[1], ary[2], ary[3])
    end
    return res
  end

  def table_turn(aa)  # 縦・横逆転した２次元配列を返す
    unless aa.class == Array
      return false
    end
    if aa[0].class != Array
      aa = [aa]
    end
    i_size = aa.size-1
    j_size = 0
    for i in 0..i_size
      next  unless aa[i].class == Array
      if (aa[i].size-1) > j_size
        j_size = aa[i].size-1
      end
    end
    if j_size == 0
      return false
    end
    r = []
    for j in 0..j_size
      w = []
      for i in 0..i_size
        w << aa[i][j]
      end
      r << w
    end
    return r
  end

  def to_enc(str)  # strの文字コードを当スクリプトのものと同じにする
  ecnv = {'e'=>"e", 'euc-jp'=>"e", 's'=>"s", 'shift_jis'=>"s",
      'u'=>"w", 'utf-8'=>"w", 'windows-31j'=>"s", 'cp932'=>"s"}
    str = str.to_s
    if RUBY_VERSION < "1.9"
      c = $KCODE.to_s.downcase[0,1]
      if c != "" and ecnv[c]
        c = ecnv[c]
        str = NKF.nkf("-m0#{c}", str)
      end
    else
      etype = 'a'.encoding.to_s  # 当スクリプトの文字コード
      c = ecnv[etype.downcase]
      str = NKF.nkf("-m0#{c}", str)  if c != nil
      str.force_encoding(etype)
    end
    return str
  end

  def file_read(fname)
    begin
      body = File.open(fullpath(fname)) {|ff|  ff.read}
    rescue
      body = ''
      STDERR.puts "can't open '#{fname}'"
    end
    if body != ''
      body = to_enc(body)
    end
    return body
  end

  def time2str(x)
    if x.class == Time
      x = x.to_s.sub(/ +[\+\-]\d+$/, "")
      x = x.sub(/( +|\:)0+$/, "") while x =~ /( +|\:)0+$/
      x = x + ":00"  if x =~ / +\d+$/
    elsif x.class == Array
      for i in 0...x.size
        x[i] = time2str(x[i])
      end
    elsif x.class == Hash
      hs = {}
      x.each {|key, val|
        hs[time2str(key)] = time2str(val)
      }
      x.clear
      x = hs
    end
    return x
  end

  def to_num(x)
    if x.class == String
      if x =~ /^-?\d+$/
        x = x.to_i
      elsif x =~ /^(-?\d+)\.0+$/
        x = $1.to_i
      elsif x =~ /^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
        x = x.to_f
      elsif x =~ /^\d{4}[\/\-]\d\d?[\/\-]\d\d?( +\d\d?:\d\d?(:\d\d?)?)?$/ or
        x =~ /^\d{4}[\/\-]\d\d?[\/\-]\d\d? +\d\d?:\d\d?:\d\d?( +[\+\-]?\d+)?$/
        x = Time.parse(x)
      end
    elsif x.class == Array
      for i in 0...x.size
        x[i] = to_num(x[i])
      end
    elsif x.class == Hash
      hs = {}
      x.each {|key, val|
        hs[to_num(key)] = to_num(val)
      }
      x.clear
      x = hs
    end
    return x
  end

  def get_sep(str)  # 区切り文字の判定
    cct = tct = 0
    if str.class == String
      str.gsub(/,/) {|ptn|
        cct += 1
        ptn = ''
      }
      str.gsub(/\t/) {|ptn|
        tct += 1
        ptn = ''
      }
    end
    return (tct > cct) ? "\t" : ","
  end

  def str2ary(str, sep = nil, atype = :num)  # 文字列を配列に変換
    sep = get_sep(str)  if sep == nil
    ary = (RUBY_VERSION < "1.9") ? CSV.parse(str, sep) :
        CSV.parse(str, {:col_sep=>sep})
    ary.map! {|row|
      row.map! {|x|
        x = x.to_s
        if x =~ /^\s*$/
          x = nil
        end
        x
      }
      row.pop  while row.size>0 and row.last==nil
      row
    }
    ary = to_num(ary)  if atype == :num
    return ary
  end

  def ary2str(aa, sep=',')  # 配列を文字列に変換
    aa = time2str(aa)
    aa.map! {|row|
      row.map! {|x|
        if x.to_s =~ /^\s*$/
          x = nil
        end
        x
      }
    }
    str = ''
    aa.each {|row|
      row = [row]  if row.class != Array
      line = (RUBY_VERSION < "1.9") ? CSV.generate_line(row, sep) :
          CSV.generate_line(row, {:col_sep=>sep})
      line.chomp!
      str += "#{line}\n"
    }
    return str
  end

  def to_utf(x)
    return to_u(x)
  end
end


# ソルバー用のクラス
class ExlSolver
    # クラス内で用いる定数
  Relations = {'<='=>1, '='=>2, '=='=>2, '>='=>3,
        'int'=>4, 'bin'=>5, 'dif'=>6}
  MaxMinVal = {'max'=>1, 'min'=>2, 'val'=>3}
  OkKey = %w(SetCell MaxMinVal ValueOf ByChange Engine EngineDesc)
  AddKey = %w(CellRef Relation FormulaText)
  OptionsKey = %w(MaxTime Iterations Precision AssumeLinear StepThru) +
      %w(Estimates Derivatives SearchOption IntTolerance Scaling) +
      %w(Convergence AssumeNonNeg PopulationSize RandomSeed) +
      %w(MultiStart RequireBounds MutationRate MaxSubproblems) +
      %w(MaxIntegerSols SolveWithout MaxTimeNoImp)
  GetNum = {'SetCell'=>1, 'MaxMinVal'=>2,
      'ValueOf'=>3, 'ByChange'=>4,
      'AddCount'=>5, 'AddLeft'=>6, 'AddRelation'=>7, 'AddRight'=>8,
      'MaxTime'=>9, 'Iterations'=>10, 'Precision'=>11,
      'AssumeLinear'=>13, 'StepThru'=>14, 'Estimates'=>16, 'Derivatives'=>17,
      'SearchOption'=>18, 'IntTolerance'=>12, 'Scaling'=>15,
      'Convergence'=>19, 'AssumeNonNeg'=>20, 'PopulationSize'=>21,
      'RandomSeed'=>22, 'MultiStart'=>23, 'RequireBounds'=>24,
      'MutationRate'=>25, 'MaxSubproblems'=>26, 'MaxIntegerSols'=>27,
      'SolveWithout'=>28, 'MaxTimeNoImp'=>29}
  SettleDefault = [["SolverSolve", true], ["SolverFinish", 1]]
  RunError = -2146826273
  OtherMethod = {'change'=>"SolverChange", 'delete'=>"SolverDelete",
      'finish'=>"SolverFinish", 'finishdialog'=>"SolverFinishDialog",
      'load'=>"SolverLoad", 'okdialog'=>"SolverOkDialog",
      'reset'=>"SolverReset", 'save'=>"SolverSave", 'solve'=>"SolverSolve"}

  def initialize(app=nil, solver_file=nil, error_out=0)
    @app = nil
    @solver_file = nil
    @auto_open = nil
    if app
      set_app(app)
      set_file(solver_file)
    end
    @error_out = error_out
  end
  attr_reader :solver_file
  attr_accessor :error_out

  def set_app(app)
    if !@app and app
      app = app.obj  if app.class == Exlap
      @app = app
      auto_open()
    end
  end

  def  set_file(solver_file = nil)
    if @app
      if !solver_file or !test(?e, solver_file)
        solver_file = nil
        @app.AddIns.each do |target|
          if target.Name =~ /^solver\.x/i
            solver_file = target.FullName
            break
          end
        end
      end
      if solver_file
        @solver_file = solver_file
      end
    end
    if !@solver_file
      STDERR.puts "can not find the addin (SOLVER.XLA | SOLVER.XLAM)"
    else
      auto_open()
    end
  end

  def auto_open()
    if !@auto_open and @app and @solver_file and
       @app.Version.to_f < 12.0
      self.run "Auto_Open"
      @auto_open = true
    end
  end

  def run(cmd, *arg)
    if !@solver_file or !@app
      return false
    end
    slv_cmd = "'#{@solver_file}'!#{cmd}"
    res = nil
    if arg.size == 0
      res = @app.Run slv_cmd
    else
      res = @app.Run(slv_cmd, *arg)
    end
    if res and @error_out != 0
      ary = [cmd]
      arg.each {|x|
        if x.class == WIN32OLE and (x2 = Exl::a1(x)) != nil
          x = x2
        end
        ary << x
      }
      str = ary.join(', ')
      str = str.gsub(/\#<WIN32OLE[^>]+>/, "OLE_object")
      out_str = res<0 ? sprintf("%d(%d): %s\n", res, res+0x7ff60000, str) :
          sprintf("%d: %s\n", res, str)
      if @error_out < 0
        STDERR.puts out_str  if res < 0
      else
        STDERR.puts out_str
      end
    end
    return res
  end

  def bat(ary)
    if ary.class != Array or ary.size < 1
      return nil
    end
    count = 0
    ary.each {|row|
      count += 1  if row.class == Array
    }
    if count == 0  # ２次元配列ではない
      res = self.run(*ary)
      return res
    end
    res = []
    ary.each {|row|
      if row.class != Array or row.size < 1
        res << nil
        next
      end
      r = self.run(*row)
      break  if r == false
      res << r
    }
    return res
  end

  def settle(arg = nil)  # Solve & Finish 連続実行
    if arg == nil
      arg = SettleDefault
    end
    self.bat(arg)
  end

  def options_key()
    return (@app.Version.to_f <= 12.0) ? OptionsKey[0..11] : OptionsKey
  end

  def get_num()
    return GetNum
  end

  def relations()
    return Relations
  end

  def options(*prm)
    if prm.size == 1 and prm[0].class == Hash
      hs = prm[0].dup
      if !hs['SearchOption'] and hs['Search']
        hs['SearchOption'] = hs['Search']
      end
      arg = ["SolverOptions"]
      self.options_key.each do |key|
        arg << hs[key]
      end
      arg.pop  while arg.size>0 and arg.last==nil
    else
      arg = ["SolverOptions"] + prm
    end
    res = nil
    if arg.size > 1
      res = self.run(*arg)
    end
    return res
  end

  def ok(*prm)
    if prm.size == 1 and prm[0].class == Hash
      arg = []
      OkKey.each do |key|
        arg << prm[0][key]
      end
      arg.pop  while arg.size>0 and arg.last==nil
    else
      arg = prm
    end
    if arg[1].class == String
      n = MaxMinVal[arg[1].downcase]
      if n
        arg[1] = n
      end
    end
    self.run("SolverOk", *arg)
  end

  def add(*prm)
    if prm.size == 1 and prm[0].class == Hash
      arg = []
      AddKey.each do |key|
        arg << prm[0][key]
      end
      arg.pop  while arg.size>0 and arg.last==nil
    else
      arg = prm
    end
    if arg[1].class == String
      n = Relations[arg[1].downcase]
      if n
        arg[1] = n
      end
    end
    self.run("SolverAdd", *arg)
  end

  def get(*prm)
    arg = prm.dup
    if arg[0].class == String and (n = GetNum[arg[0]])
      arg[0] = n
    end
    return self.run("SolverGet", *arg)
  end

  def get_options()
    hs = {}
    self.options_key.each {|key|
      num = GetNum[key]
      val = self.get(num)
      next  if val == RunError
      hs[key] = val
    }
    return hs
  end

  def method_missing(sym, *arg)
    name = sym.to_s
    mtd = OtherMethod[name.downcase]
    unless mtd
      STDERR.puts "undefined method `#{name}' for ExlSolver (NameError)"
      return
    end
    return self.run(mtd, *arg)
  end
end

include ExlConstant
