# rpt.rb: (encoding: Windows-31J)  2014/09/11
  # Copyright (C) T. Yoshiizumi, 2014 All rights reserved.

if !ENV['RPT_TMP']
  STDERR.puts "'Rpt' module is available on condition that " +
      "the RPT package has been installed."
else
  require "rrxwin"
  require "exlap"
  require "wrdap"
  require "yado"

  ## main methods
  module Rpt
    module_function()

    def kramdown_tex_s(str, *opt)
      str = Rrx::kramdown_tex(str, :s, *opt)
      if str.class == String
        # str = NKF.nkf("-m0s", str)
        if str =~ /^\\usepackage\{graphicx\}/
          str = str.sub(/^\\usepackage\{graphicx\}/,
              "\\usepackage[dvipdfmx]{graphicx}")
        elsif str =~ /^\\usepackage\[(.*?)\]\{graphicx\}/
          str = str.sub(/^\\usepackage\[(.*?)\]\{graphicx\}/) do |ptn|
            opt = $1
            if opt =~ /^\s*$/
              opt = 'dvipdfmx'
            elsif opt !~ /dvipdfmx/i
              opt = opt + ',dvipdfmx'
            end
            ptn = "\\usepackage[#{opt}]{graphicx}"
          end
        end
      end
      return str
    end

    def pandoc_tex_s(str, *opt_ary)
      str = Rrx::pandoc_tex(str, :s, *opt_ary)
      if str.class == String
        str = NKF.nkf("-m0s", str)
        if str =~ /^\\usepackage\{graphicx\}/
          str = str.sub(/^\\usepackage\{graphicx\}/,
              "\\usepackage[dvipdfmx]{graphicx}")
        elsif str =~ /^\\usepackage\[(.*?)\]\{graphicx\}/
          str = str.sub(/^\\usepackage\[(.*?)\]\{graphicx\}/) do |ptn|
            opt = $1
            if opt =~ /^\s*$/
              opt = 'dvipdfmx'
            elsif opt !~ /dvipdfmx/i
              opt = opt + ',dvipdfmx'
            end
            ptn = "\\usepackage[#{opt}]{graphicx}"
          end
        end
      end
      return str
    end

    def tex2pdf(tex_str)
      if tex_str.class != String or tex_str == ''
        return nil
      end
      tmpdir = ENV['RPT_TMP'].gsub(/\\/, "/")
      tex_file = pref = nil
      nn = 0
      while tex_file == nil
        pref = sprintf("$$RPT%04d_", nn)
        if !test(?e, "#{pref}tmp.tex")
          tex_file = "#{pref}tmp.tex"
        end
        nn += 1
      end
      Dir.glob("#{pref}*") {|fname|  File.unlink(fname)}
      File.open(tex_file, "w") {|ff|  ff.write tex_str}
      out_file = "#{tmpdir}/#{pref}out.txt"
      File.unlink(out_file)  if test(?e, out_file)
      stdout_old = $stdout.dup
      $stdout.reopen(out_file)
      err_file = "#{tmpdir}/#{pref}err.txt"
      File.unlink(err_file)  if test(?e, err_file)
      stderr_old = $stderr.dup
      $stderr.reopen(err_file)
      cmd = "platex  \"#{tex_file}\""
      system(cmd)
      if test(?e, "#{pref}tmp.toc")
        system(cmd)
      end
      if test(?e, "#{pref}tmp.dvi")
        cmd = "dvipdfmx \"#{pref}tmp.dvi\""
        system(cmd)
      end
      $stdout.flush
      $stderr.flush
      $stdout.reopen stdout_old
      $stderr.reopen stderr_old
      $X_OUT = $X_ERROR = ''
      if test(?e, out_file)
        $X_OUT = File.read(out_file)  if !test(?z, out_file)
        File.unlink(out_file)
      end
      if test(?e, err_file)
        $X_ERROR = File.read(err_file)  if !test(?z, err_file)
        File.unlink(err_file)
      end
      pdf_file = tex_file.sub(/\.tex$/i, ".pdf")
      pdf_str = nil
      if test(?e, pdf_file) and !test(?z, pdf_file)
        File.open(pdf_file, "rb") {|ff|  pdf_str = ff.read}
      end
      Dir.glob("#{pref}*") {|fname|  File.unlink(fname)}
      Dir.glob("*.xbb") {|fname|  File.unlink(fname)}
      if !pdf_str
        File.open("X_OUT.TXT", "w") {|ff|  ff.write $X_OUT}
        File.open("X_ERROR.TXT", "w") {|ff|  ff.write $X_ERROR}
      end
      return pdf_str
    end

    def kramdown_pdf(str, *opt)
      tex_str = kramdown_tex_s(str, *opt)
      return tex2pdf(tex_str)
    end

    def pandoc_pdf(str, *opt)
      tex_str = pandoc_tex_s(str, *opt)
      return tex2pdf(tex_str)
    end
  end
end
