# rrxwin.rb version 1.12: (coding: Windows-31J)  2018/04/04
  # a ruby library for R(statistical analysis command)
  # on the clipboard operation, Cf. the web site below
  # http://tailcodenote.shisyou.com/code/ruby/006.html
  # Copyright (C) T. Yoshiizumi, 2013-2018 All rights reserved.
require "tempfile"
require "csv"
require "nkf"
require "open-uri"
require "time"
require "erb"
if RUBY_PLATFORM =~ /mswin(?!ce)|mingw|cygwin|bccwin/i
  require "win32ole"
end
begin
  require "kramdown"
rescue LoadError
end
begin
  require "nokogiri"
rescue LoadError
end
begin
  require "terminal-table"
rescue LoadError
end
begin
  require "pandoc-ruby"
rescue LoadError
end
if defined?(Roo)
  begin
    require "roo-xls"
  rescue LoadError
  end
  begin
    require "roo-google"
  rescue LoadError
  end
end

# module for clipboard of MS-Windows
if defined?(WIN32OLE)
  if RUBY_VERSION < "2.0"
    require "Win32API"
    if RUBY_VERSION.to_f >= 1.9
      require "dl"
    end
  
    module Rclip
      @@OpenClipboard = Win32API.new('user32', 'OpenClipboard', ['I'], 'I')
      @@CloseClipboard = Win32API.new('user32', 'CloseClipboard', [], 'I')
      @@GetClipboardData = Win32API.new('user32',
          'GetClipboardData', ['I'], 'I')
      @@SetClipboardData = Win32API.new('user32',
          'SetClipboardData', ['I', 'I'], 'I')
      @@EmptyClipboard = Win32API.new('user32', 'EmptyClipboard', [], 'I')
      @@EnumClipboardFormats = Win32API.new('user32',
          'EnumClipboardFormats', ['I'], 'I')
  
      @@GlobalAlloc = Win32API.new('kernel32', 'GlobalAlloc', ['I','I'], 'I')
      @@GlobalLock = Win32API.new('kernel32', 'GlobalLock', ['I'], 'P');   #get
      @@GlobalLockI = Win32API.new('kernel32', 'GlobalLock', ['I'], 'I');  #set
      @@GlobalUnlock = Win32API.new('kernel32', 'GlobalUnlock', ['I'], 'I')
      @@GlobalSize = Win32API.new('kernel32', 'GlobalSize', ['I'], 'I')
      @@GlobalFree = Win32API.new('kernel32', 'GlobalFree', ['I'], 'I')
      if RUBY_VERSION.to_f >= 1.9
        begin
          @@memcpy = Win32API.new('msvcrt', 'memcpy', ['I', 'P', 'I'], 'I')
        rescue
          STDERR.puts "need 'msvcrt.dll' for copying data to Clipboard."
        end
      end
  
      @@lstrcpy = Win32API.new('kernel32', 'lstrcpyA', ['P', 'P'], 'P')
      @@lstrlen = Win32API.new('kernel32', 'lstrlenA', ['P'], 'I')
  
      CF = {:CF_TEXT=>1, :CF_BITMAP=>2, :CF_METAFILEPICT=>3, :CF_SYLK=>4,
        :CF_DIF=>5, :CF_TIFF=>6, :CF_OEMTEXT=>7, :CF_DIB=>8, :CF_PALETTE=>9,
        :CF_PENDATA=>10, :CF_RIFF=>11, :CF_WAVE=>12, :CF_UNICODETEXT=>13,
        :CF_ENHMETAFILE=>14, :CF_HDROP=>15, :CF_LOCALE=>16, :CF_DIBV5=>17,
        :CF_OWNERDISPLAY=>0x80, :CF_DSPTEXT=>0x81, :CF_DSPBITMAP=>0x82,
        :CF_DSPMETAFILEPICT=>0x83, :CF_DSPENHMETAFILE=>0x8e,
        :ObjectLink=>0xc002, :OwnerLink=>0xc003, :Native=>0xc004,
        :DataObject=>0xc009, :TravelBand=>0xc27e, :IBM_Tivoli_Win=>0xc282,
        :QueryBuilderBand=>0xc296, :TOOLBAR_CUSTOMIZE=>0xc299,
        :"Address Band Root"=>0xc27f, :"RichEdit Binary"=>0xc297,
        :"Object Descriptor"=>0xc00e, :"Preferred DropEffect"=>0xc157,
        :"WPD NSE PnPDevicePath"=>0xc281, :"Ole Private Data"=>0xc013,
        :"Embed Source"=>0xc00b, :"Link Source"=>0xc00d,
        :"Link Source Descriptor"=>0xc00f, :"Rich Text Format"=>0xc0ec}
      GM = {:GMEM_FIXED=>0, :GMEM_MOVEABLE=>2,
        :GMEM_ZEROINIT=>0x40, :GHND=>0x42}
  
      module_function()
      def clear()
        @@EmptyClipboard.Call()
      end
  
      def getData(cf = :CF_TEXT)
        cf = CF[cf]  if cf.class == Symbol
        clpdata = ""
        while @@OpenClipboard.Call(0) == 0
          sleep(0.2)
        end
        begin
          if (hnd = @@GetClipboardData.Call(cf)) != 0
            if ptr = @@GlobalLock.Call(hnd)
              clpdata = (RUBY_VERSION < "1.9") ? ptr : DL::CPtr[ptr].to_s
              @@GlobalUnlock.Call(hnd)
            end
          end
        ensure
          @@CloseClipboard.Call
        end
        return clpdata
      end
  
      def setData(str, cf = :CF_TEXT, gm = :GHND)
        cf = CF[cf]  if cf.class == Symbol
        gm = GM[gm]  if gm.class == Symbol
        while @@OpenClipboard.Call(0) == 0
          sleep(0.2)
        end
        if clear() != 0
          len = @@lstrlen.Call(str)
          hmem = @@GlobalAlloc.Call(gm, len+1)
          if RUBY_VERSION < "1.9"
            pmem = @@GlobalLockI.Call(hmem)
            @@lstrcpy.Call(pmem, str)
          else
            pmem = @@GlobalLock.Call(hmem)
            @@memcpy.Call(pmem, str, len+1)
          end
          @@SetClipboardData.Call(cf, hmem)
          @@GlobalUnlock.Call(hmem)
          @@CloseClipboard.Call
        end
      end
  
      def formats()
        res = []
        while @@OpenClipboard.Call(0) == 0
          sleep(0.2)
        end
        fmt = @@EnumClipboardFormats.Call(0)
        while fmt != 0
          fs = (RUBY_VERSION < "1.9") ? CF.index(fmt) : CF.key(fmt)
          fs = fmt  unless fs
          res << fs
          fmt = @@EnumClipboardFormats.Call(fmt)
        end
        return res
      end
    end  ## end of Rclip module for ruby ver 1.8 | 1.9
  
  else  # if RUBY_VERSION >= "2.0"
    require "fiddle"
    require "fiddle/import"
  
    module Rclip
      module W
        extend Fiddle::Importer
        dlload 'user32.dll', 'kernel32.dll', 'msvcrt.dll'
        extern "int OpenClipboard(int)"
        extern "int CloseClipboard()"
        extern "int GetClipboardData(int)"
        extern "int SetClipboardData(int, int)"
        extern "int EmptyClipboard()"
        extern "int EnumClipboardFormats(int)"
        extern "int GlobalAlloc(int, int)"
        extern "char* GlobalLock(int)"
        extern "int GlobalLock(int)"
        extern "int GlobalUnlock(int)"
        extern "int GlobalSize(int)"
        extern "int GlobalFree(int)"
        extern "int memcpy(int, char*, int)"
        extern "char* lstrcpyA(char*, char*)"
        extern "int lstrlenA(char*)"
      end
  
      CF = {:CF_TEXT=>1, :CF_BITMAP=>2, :CF_METAFILEPICT=>3, :CF_SYLK=>4,
        :CF_DIF=>5, :CF_TIFF=>6, :CF_OEMTEXT=>7, :CF_DIB=>8, :CF_PALETTE=>9,
        :CF_PENDATA=>10, :CF_RIFF=>11, :CF_WAVE=>12, :CF_UNICODETEXT=>13,
        :CF_ENHMETAFILE=>14, :CF_HDROP=>15, :CF_LOCALE=>16, :CF_DIBV5=>17,
        :CF_OWNERDISPLAY=>0x80, :CF_DSPTEXT=>0x81, :CF_DSPBITMAP=>0x82,
        :CF_DSPMETAFILEPICT=>0x83, :CF_DSPENHMETAFILE=>0x8e,
        :ObjectLink=>0xc002, :OwnerLink=>0xc003, :Native=>0xc004,
        :DataObject=>0xc009, :TravelBand=>0xc27e, :IBM_Tivoli_Win=>0xc282,
        :QueryBuilderBand=>0xc296, :TOOLBAR_CUSTOMIZE=>0xc299,
        :"Address Band Root"=>0xc27f, :"RichEdit Binary"=>0xc297,
        :"Object Descriptor"=>0xc00e, :"Preferred DropEffect"=>0xc157,
        :"WPD NSE PnPDevicePath"=>0xc281, :"Ole Private Data"=>0xc013,
        :"Embed Source"=>0xc00b, :"Link Source"=>0xc00d,
        :"Link Source Descriptor"=>0xc00f, :"Rich Text Format"=>0xc0ec}
      GM = {:GMEM_FIXED=>0, :GMEM_MOVEABLE=>2,
        :GMEM_ZEROINIT=>0x40, :GHND=>0x42}
  
      module_function()
      def clear()
        W.EmptyClipboard()
      end
  
      def getData(cf = :CF_TEXT)
        cf = CF[cf]  if cf.class == Symbol
        clpdata = ""
        while W.OpenClipboard(0) == 0
          sleep(0.2)
        end
        begin
          if (hnd = W.GetClipboardData(cf)) != 0
            if (ptr = W.GlobalLock(hnd))
              clpdata = Fiddle::Pointer[ptr].to_s
              W.GlobalUnlock(hnd)
            end
          end
        ensure
          W.CloseClipboard()
        end
        return clpdata
      end
  
      def setData(str, cf = :CF_TEXT, gm = :GHND)
        cf = CF[cf]  if cf.class == Symbol
        gm = GM[gm]  if gm.class == Symbol
        while W.OpenClipboard(0) == 0
          sleep(0.2)
        end
        if clear() != 0
          len = W.lstrlenA(str)
          hmem = W.GlobalAlloc(gm, len+1)
          pmem = W.GlobalLock(hmem)
          W.memcpy(pmem, Fiddle::Pointer[str], len+1)
          W.SetClipboardData(cf, hmem)
          W.GlobalUnlock(hmem)
          W.CloseClipboard()
        end
      end
  
      def formats()
        res = []
        while W.OpenClipboard(0) == 0
          sleep(0.2)
        end
        fmt = W.EnumClipboardFormats(0)
        while fmt != 0
          fs = (RUBY_VERSION < "1.9") ? CF.index(fmt) : CF.key(fmt)
          fs = fmt  unless fs
          res << fs
          fmt = W.EnumClipboardFormats(fmt)
        end
        return res
      end
    end
  end  ## end of Rclip module for ruby ver 2.0
end


# sub_methods for Rrx
module RrxSub
  module_function()
  def get_sep(str)  # ؂蕶̔
    cct = tct = 0
    if str.class == String
      str.gsub(/,/) {|ptn|
        cct += 1
        ptn = ''
      }
      str.gsub(/\t/) {|ptn|
        tct += 1
        ptn = ''
      }
    end
    return (tct > cct) ? "\t" : ","
  end

  def rm_upper(ary)  # ̋󔒍s폜
    ary.map! {|row|
      while row.size > 0 and row.last.to_s =~ /^\s*$/
        row.pop  
      end
      row
    }
    while ary.size > 0 and ary[0] == []
      ary.shift  
    end
    return ary
  end

  def mark_num(ary)  # l}[NÂ
    res = []
    for i in 0...ary.size
      row = Array.new
      for j in 0...ary[i].size
        s = ary[i][j].to_s
        if s =~ /^-?\d+$|^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
          row[j] = [i,j]
        else
          row[j] = nil
        end
      end
      res << row
    end
    return res
  end

  def mark_data(ary)  # łȂ}[NÂ
    res = []
    for i in 0...ary.size
      row = Array.new
      for j in 0...ary[i].size
        s = ary[i][j].to_s
        unless s =~ /^\s*$/
          row[j] = [i,j]
        else
          row[j] = nil
        end
      end
      res << row
    end
    return res
  end

  def square_size(ary, addr)
    y1,x1 = addr
    unless ary[y1][x1]
      return [0, nil, nil]
    end
    if !(ary[y1][x1+1] and ary[y1+1] and ary[y1+1][x1] and ary[y1+1][x1+1])
      return [1, [y1], [x1]]
    end
    d = 2
    while ary[y1+d]
      w = ary[y1+d][x1..(x1+d)]
      break  if !w or w.size < (d+1) or w.include?(nil)
      w = []
      y=y1;  x=x1+d
      while y <= (y1+d) and ary[y][x]
        w << ary[y][x]
        y += 1
      end
      break  if !w or w.size < (d+1) or w.include?(nil)
      d += 1
    end
    yy = y1+d-1
    xx = x1+d-1
    y = yy+1
    while ary[y]
      w = ary[y][x1..xx]
      break  if !w or w.size < d or w.include?(nil)
      y += 1
    end
    y2 = y-1
    res1 = [d, (y1..y2).to_a, (x1..xx).to_a]
    res1_v = (y2-y1+1) * (xx-x1+1)
    x = xx+1
    while ary[y1][x]
      w = [ary[y1][x]]
      for y in (y1+1)..yy
        w << ary[y][x]
      end
      break  if !w or w.size < d or w.include?(nil)
      x += 1
    end
    x2 = x-1
    res2 = [d, (y1..yy).to_a, (x1..x2).to_a]
    res2_v = (yy-y1+1) * (x2-x1+1)
    return (res2_v > res1_v) ? res2 : res1
  end

  def grouping_box(boxes)
    res = []
    del_count = 0
    while box = boxes.shift
      d,yy,xx = box
      i = boxes.size-1
      while i >= 0
        d2,yy2,xx2 = boxes[i]
        if (yy & yy2).size > 0 and (xx & xx2).size > 0
          boxes.delete_at(i)
          del_count += 1
          if d2 > d
            d,yy,xx = [d2,yy2,xx2]
          elsif d2 == d and yy2.size*xx2.size > yy.size*xx.size
            d,yy,xx = [d2,yy2,xx2]
          end
        end
        i -= 1
      end
      res << [d,yy,xx]
    end
    if del_count > 0
      res = grouping_box(res)
    end
    return res
  end

  def get_boxes(ary)
    boxes = []
    ary.each_with_index {|row, i|
      next  if row.class != Array  or row.size < 1
      row.each_with_index {|cell, j|
        next  unless cell
        d,yy,xx = square_size(ary, [i,j])
        next  if d < 2
        boxes << [d,yy,xx]
      }
    }
    if boxes.size > 1
      boxes = grouping_box(boxes)
    end
    return boxes
  end

  def each_square(ary, &block)  # eX̍s̍WSǂ
    unless block
      return
    end
    boxes = get_boxes(ary)
    boxes.each {|box|
      d,yy,xx = box
      w = [yy[0], xx[0], yy[-1], xx[-1]]
      block.call(w)
    }
  end

  def select_num2(ary)  # l̒c{o̒o
    ary2 = []
    nary = RrxSub.mark_num(ary)
    RrxSub.each_square(nary) {|y1,x1, y2,x2|
      next  if !(y2 > y1 and x2 > x1)
      y0 = (y1>0) ? (y1-1) : y1
      x0 = (x1>0) ? (x1-1) : x1
      read_opt = (y1 > y0) ? "header=T" : "header=F"
      read_opt = read_opt + ", row.names=1"  if x1 > x0
      title = []
      title.push(ary[y0-1][x0-1].to_s)  if y1>y0 and y0>0 and
          x1>x0 and x0>0 and !nary[y0-1][x0-1]
      title.push(ary[y0-1][x0].to_s)  if y1>y0 and y0>0 and !nary[y0-1][x0]
      title.push(ary[y0][x0-1].to_s)  if x1>x0 and x0>0 and !nary[y0][x0-1]
      title.delete_if {|x|  x =~ /^\s*$/}
      wary = []
      for i in y0..y2
        row = Array.new
        for j in x0..x2
          val = ary[i][j]
          if val.to_s =~ /^\s*$/ and [i,j] != [y1-1,x1-1]
            if i == (y1-1)
              val = sprintf("x%d", j-((x1==x0) ? (x0-1) : x0))
            end
            if j == (x1-1)
              val = sprintf("y%d", i-((y1==y0) ? (y0-1) : y0))
            end
          end
          row << val
        end
        wary << row
      end
      ary2 << [wary, title.join("|"), read_opt]
    }
    return ary2
  end

  def select_data2(ary)
    ary2 = []
    sary = RrxSub.mark_data(ary)
    RrxSub.each_square(sary) {|y1,x1, y2,x2|
      next  if !(y2 > y1 and x2 > x1)
      num_regex = /^-?\d+$|^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
      num_flag = false
      for j in x1..x2
        if ary[y1][j].to_s =~ num_regex
          num_flag = true
          break
        end
      end
      read_opt = num_flag ? "header=F" : "header=T"
      num_flag = false
      for i in y1..y2
        if ary[i][x1].to_s =~ num_regex
          num_flag = true
          break
        end
      end
      read_opt += ", row.names=1"  unless num_flag
      title = []
      title.push(ary[y1-1][x1-1].to_s)  if y1>0 and x1>0
      title.push(ary[y1-1][x1].to_s)  if y1>0
      title.push(ary[y1][x1-1].to_s)  if x1>0
      title.delete_if {|x|  x =~ /^\s*$/}
      wary = []
      for i in y1..y2
        row = Array.new
        for j in x1..x2
          row << ary[i][j]
        end
        wary << row
      end
      ary2 << [wary, title.join("|"), read_opt]
    }
    return ary2
  end

  def select_squares(ary, type)  # l̒c̒o
    ary2 = []
    nary = (type == :num) ? RrxSub.mark_num(ary) : RrxSub.mark_data(ary)
    RrxSub.each_square(nary) {|y1,x1, y2,x2|
      if y2 > y1 and x2 > x1
        wary = []
        for i in y1..y2
          row = Array.new
          for j in x1..x2
            row << ary[i][j]
          end
          wary << row
        end
        ary2 << wary
      end
    }
    return ary2
  end

  def table_turn(aa)  # cEt]QzԂ
    unless aa.class == Array
      return false
    end
    if aa[0].class != Array
      aa = [aa]
    end
    i_size = aa.size-1
    j_size = 0
    for i in 0..i_size
      next  unless aa[i].class == Array
      if (aa[i].size-1) > j_size
        j_size = aa[i].size-1
      end
    end
    if j_size == 0
      return false
    end
    r = []
    for j in 0..j_size
      w = []
      for i in 0..i_size
        w << aa[i][j]
      end
      r << w
    end
    return r
  end

  def temp_prefix()
    ff = Tempfile.new("ww")
    ff.close
    res = File.dirname(ff.path)
    ff.unlink
    return res
  end

  def mshttp_obj(os_type = nil)
    if !os_type
      os_type = defined?(WIN32OLE) ? :WIN : :COMMON
    end
    mshttp = nil
    if os_type != :WIN
      return mshttp
    end
    list = %w(Msxml2.XMLHTTP Msxml.XMLHTTPRequest Microsoft.XMLHTTP)
    list.each do |x_name|
      begin
        mshttp = WIN32OLE.new(x_name)
      rescue
        mshttp = nil
      end
      if mshttp != nil
        break
      end
    end
    return mshttp
  end
end

# main methods
module Rrx
  @@rrx_cfg = sprintf("%s/rrx.cfg", File.dirname(
      defined?(ExerbRuntime) ? ExerbRuntime.filepath :
      (ENV['OCRA_EXECUTABLE'] ? ENV['OCRA_EXECUTABLE'] : __FILE__)))
  @@path = nil  # R̃tpX
  @@opt = nil
  @@log = nil
  @@encopt = true
  @@prefix = "rrx_"  # e|t@C̓
  @@file_hs = {}
  @@temp_hs = {}
  @@temp_prefix = RrxSub.temp_prefix()
  @@ext_regex = /\.(xls|xml|xlsx|xlsm|ods)$/i
  @@charset = nil
  @@r_result = nil
  @@r_keys = nil
  @@robj_list = []
  @@http_opt = nil

  NCNV = {'e'=>"-e", 'euc-jp'=>"-e", 's'=>"--sjis", 'shift_jis'=>"-s",
      'u'=>"-w", 'utf-8'=>"-w", 'windows-31j'=>"--sjis", 'cp932'=>"--sjis"}
  FS = "\001\001"
  RS = "\002\002"
  VFS = "\005\005"
  USS = "\x13\x13"
  UTT = "\x14\x14"
  CSV_MARK = "!!csv"
  @@robj = "robj.count <- 0\n" +
    "robj.idx <- 0\n" +
    "robj.except <- c(\"symbol\",\"pairlist\",\"promise\",\"language\"," +
    "\"char\",\"...\",\"any\",\"expression\",\"externalptr\",\"bytecode\"," +
    "\"weakref\")\n" +
    "robj.out <- function(item, tt, outfile) {\n" +
    "\tif (is.null(dim(tt)) && is.table(tt) && !is.matrix(tt))\n" +
    "\t\ttt <- as.matrix(tt)\n" +
    "\tcat(\"#{USS}\", item, \"#{FS}\", file=outfile, sep=\"\")\n" +
    "\tif (is.list(tt)) robj.each(tt, outfile)\n" +
    "\telse if (!is.null(dim(tt))) {\n" +
    "\t\tcat(\"#{CSV_MARK}#{FS}\", file=outfile, sep=\"\")\n" +
    "\t\twrite.table(tt, file=outfile, sep=\",\", quote=T, col.names=NA)}\n" +
    "\telse {\n" +
    "\t\tif (length(robj.except[robj.except == typeof(tt)]) > 0) " +
    "{tt <- as.character(tt)}\n" +
    "\t\tcat(tt, file=outfile, sep=\"#{VFS}\")\n\t}\n" +
    "\tcat(\"#{RS}#{UTT}\", file=outfile, sep=\"\")\n}\n" +
    "robj.i <- function(zz = NA) {\n" +
    "\tif (is.na(zz)) robj.idx <<- robj.idx + 1\n" +
    "\telse robj.idx <<- zz\n" +
    "\treturn(robj.idx)\n" +
    "}\n" +
    "robj.each <- function(xx, outfile) {\n" +
    "\tnn <- names(xx)\n" +
    "\tif (is.null(nn)) nn <- attributes(xx)[[\"names\"]]\n" +
    "\tif (is.null(nn) && length(xx) == 1) nn <- c(\"self\")\n" +
    "\tif (is.null(nn)) {\n" +
    "\t\tv <- NULL\n" +
    "\t\tfor (j in 1:length(xx))\n" +
    "\t\t\tv <- append(v, sprintf(\"_%d_\", robj.i()))\n" +
    "\t\tnames(xx) <- v\n" +
    "\t\tnn <- names(xx)\n" +
    "\t}\n" +
    "\tfor (i in 1:length(nn)) {\n" +
    "\t\tif (length(nn[i]) == 0) next\n" +
    "\t\titem <- nn[i]\n" +
    "\t\tif (is.na(item) || is.nan(item) || is.null(item)) next\n" +
    "\t\tif (item == \"\" && length(nn) == 1)  item <- \"self\"\n" +
    "\t\telse if (item == \"\")  item <- sprintf(\"_%d_\",robj.i())\n" +
    "\t\trobj.out(item, xx[[i]], outfile)\n" +
    "\t}\n" +
    "}\n" +
    "robj <- function(xx=NULL, kk=NA) {\n" +
    "\tif (is.null(xx))  return(invisible(xx))\n" +
    "\tif (is.na(kk)) {kk <- robj.count + 1; robj.count <<- kk;\n" +
    "\t\tfname <- sprintf(\"#{@@temp_prefix}/ROBJ_robj%.0f.rrx\",kk)}\n" +
    "\telse  {fname <- sprintf(\"#{@@temp_prefix}/ROBJ_%s.rrx\",kk)}\n" +
    "\toutfile <- file(fname, open=\"w\")\n" +
    "\trobj.i(0)\n" +
    "\tnn <- names(xx)\n" +
    "\tif (is.null(nn)) nn <- attributes(xx)[[\"names\"]]\n" +
    "\tif (is.null(nn)) {\n" +
    "\t\tif (is.list(xx)) robj.each(xx, outfile)\n" +
    "\t\telse robj.out(\"self\", xx, outfile)\n" +
    "\t}\n" +
    "\telse if (!is.null(dim(xx)) && !is.list(xx))\n" +
        "\t\trobj.out(\"self\", xx, outfile)\n" +
    "\telse robj.each(xx, outfile)\n" +
    "\tclose(outfile)\n" +
    "\treturn(invisible(xx))\n" +
    "}\n"

  module_function()
  def path=(val)
    old_path = @@path
    if val.class == String
      @@path = val
      unless old_path
        hs = rload()
        hs[:path] = @@path
        rsave(hs)
      end
    end
    @@path
  end
  def path()
    if !@@path and test(?f, @@rrx_cfg)
      hs = rload()
      @@path = hs[:path]
    end
    if !@@path
      wpath = search_rcmd_path()
      if wpath != nil
        @@path = wpath
      else
        @@path = defined?(WIN32OLE) ? "Rcmd" : "R"
      end
    end
    @@path
  end
  def cfg_delete()
    if test(?e, @@rrx_cfg)
      File.unlink(@@rrx_cfg)
    end
  end
  def opt=(val)
    @@opt = val
  end
  def opt()
    @@opt
  end
  def log=(val)
    @@log = val
  end
  def log()
    @@log
  end
  def encopt=(val)
    @@encopt = val
  end
  def encopt()
    @@encopt
  end

  def rencode(body)
    if body.class != String or body == ''
      return body
    end
    if @@charset != nil
      code = @@charset
      c = NCNV[code.downcase]
      if c != nil
        body = NKF.nkf("-m0 #{c}", body)
        if RUBY_VERSION.to_f >= 1.9
          body.force_encoding(code)
        end
      end
    elsif RUBY_VERSION < "1.9"
      c = $KCODE.to_s.downcase[0,1]
      if c != "" and NCNV[c]
        c = NCNV[c]
        body = NKF.nkf("-m0 #{c}", body)
      end
    else
      etype = __ENCODING__.to_s  # XNvg̕R[h
      c = NCNV[etype.downcase]
      body = NKF.nkf("-m0 #{c}", body)  if c != nil
      body.force_encoding(etype)
    end
    return body
  end

  def file_read(fname)
    mode = (RUBY_VERSION < "1.9") ? "r" : "r:ASCII-8BIT"
    begin
      body = File.open(fname, mode) {|ff|  ff.read}
    rescue
      body = ''
    end
    return rencode(body)
  end

  def rexec(script, logname=nil, options=nil)  # RvO̎s
    rload(:set)
    @@charset = (script.class == String) ? charset(script) : nil
    res = {}
    @@r_keys = []
    if script !~ /\r|\n/ and test(?f, script)
      script = file_read(script)
    end
    if !script or script == ''
      STDERR.puts "the R program is empty."
      return res
    end
    Dir.glob("#{@@temp_prefix}/ROBJ_*.rrx") {|fname|  File.unlink(fname)}
    gv = global_variables.dup
    %w($stdin $stderr $stdout $defout $deferr).each {|key| gv.delete(key)}
    script2 = script.dup
    while script2 =~ /\@\@\$[a-z].*?\@\@/
      script2 = script2.gsub(/\@\@(\$[a-z].*?)\@\@/) {|ptn|
        kk = $1.dup
        if (vv = eval(kk)) != nil
          ptn = "#{vv}"
        else
          STDERR.puts "can't replace '@@#{kk}@@' in method 'rexec'"
          ptn = "??"
        end
      }
    end
    tf = []
    script2 = script2.gsub(/\@\@(.+?)\@\@/) {|ptn|
      key = $1.dup
      unless @@file_hs[key]
        ext = nil
        if key =~ /^.+(\.[a-z0-9]+)$/i
          ext = $1.dup
        end
        tt = @@file_hs[key] = temp_make(ext, false)
        tf << tt
      end
      ptn = @@file_hs[key].path
    }
    temp_delete(@@file_hs[:sink])  if @@file_hs[:sink]
    tt = @@file_hs[:sink] = temp_make(false)
    tf << tt
    prescript = "#{@@robj}"
    if @@encopt and @@charset != nil
      cset = (@@charset =~ /Windows-31J/i) ? 'cp932' : @@charset
      prescript = prescript + "options(encoding=\"#{cset}\")\n"
    end
    wdd = ENV['RRX_SOURCE']
    wdd = ENV['RPT_WITH']  unless wdd
    if wdd and test(?d, wdd)
      wdd = wdd.gsub(/\\/, "/")
      wdd = wdd.sub(/\/$/, "")
      Dir.glob("#{wdd}/*.[Rr]") do |wrr|
        prescript = prescript + "source(#{wrr.inspect})\n"
      end
    end
    prescript = prescript + "sink(file=\"#{@@file_hs[:sink].path}\", " +
        "split=T) ## rrx_sink\n"
    script2 = prescript + script2
    temp_delete(@@file_hs[:script])  if @@file_hs[:script]
    tt = @@file_hs[:script] = temp_make(script2, false)
    tf << tt
    lff = temp_make(false)
    tf << lff
    if options != nil
      opt = options
    else
      wopt = defined?(WIN32OLE) ? "--no-save -q --internet2" : "--no-save -q"
      opt = !@@opt ? wopt : @@opt
    end
    unless opt =~ /--save|--no-save|--vanilla/i
      opt = opt + " --no-save"
    end
    if @@encopt and opt !~ /--encoding/i and @@charset != nil
      cset = (@@charset =~ /Windows-31J/i) ? 'cp932' : @@charset
      opt = opt + " --encoding=#{cset}"
    end
    cmd1 = defined?(WIN32OLE) ? "\"#{path()}\"" : "\"#{path()}\" CMD"
    cmd = "#{cmd1} BATCH #{opt} " +
        "\"#{@@file_hs[:script].path}\" \"#{lff.path}\""
    sres = system( cmd )
    logname = @@log  unless logname
    unless sres
      cmd2 = "#{cmd1} BATCH #{opt} R_program"
      cmd2 += " \"#{logname}\""  if logname
      STDERR.print "** failed to execute the command\n\t'#{cmd2}'\n\n"
    end
    if test(?z, lff.path)
      tf.delete(lff)
      temp_delete(lff)
    elsif test(?f, lff.path)
      body = rencode( temp_read(lff) )
      body = body.sub(/\A(.+? ## rrx_sink)\n(>.+)\Z/m, "\\2")
      res[:log] = body
      @@r_keys << :log
      tf.delete(lff)
      temp_delete(lff)
      if logname
        File.open(logname, "w") {|ff|  ff.print body}
      end
      if !sres
        STDERR.print "** log:\n" + res[:log] + "\n"
      end
    end
    @@file_hs.each {|key, ff|
      next  if !test(?f, ff.path) or test(?z, ff.path)
      res[key] = rencode( temp_read(ff) )
      @@r_keys << key
      if key == :sink
        res[key] = res[key].sub(/^> proc\.time\([^>]+\Z/, "")
        res[key] = res[key].gsub(/ +\n/, "\n")
      end
    }
    while tt = tf.shift
      temp_delete(tt)
    end
    @@robj_list = []
    Dir.glob("#{@@temp_prefix}/ROBJ_*.rrx") {|fname|
      key = nil
      if fname =~ /[\/\\]ROBJ_(.+)\.rrx$/  then  key=$1;  end
      unless key
        File.unlink(fname)
        next
      end
      val = file_read(fname)
      mtm = File.stat(fname).mtime.strftime("%Y-%m-%d %H:%M:%S")
      File.unlink(fname)
      res[key] = robj_convert(val)
      key0 = key.sub(/([^0-9])(\d+)$/) do |ptn|
        ptn = sprintf("%s%05d", $1, $2.to_i)
      end
      @@robj_list << [mtm, key0, key]
    }
    @@robj_list = @@robj_list.sort
    @@robj_list.each {|mtm, key0, key|  @@r_keys << key}
    res.delete(:script)  if res[:script]
    @@r_keys.delete(:script)
    @@r_result = res
    return res
  end

  def to_enc(str)  # str̕R[h𓖃XNvĝ̂Ɠɂ
  ecnv = {'e'=>"e", 'euc-jp'=>"e", 's'=>"s", 'shift_jis'=>"s",
      'u'=>"w", 'utf-8'=>"w", 'windows-31j'=>"s", 'cp932'=>"s"}
    str = str.to_s
    if RUBY_VERSION < "1.9"
      c = $KCODE.to_s.downcase[0,1]
      if c != "" and ecnv[c]
        c = ecnv[c]
        str = NKF.nkf("-m0#{c}", str)
      end
    else
      etype = __ENCODING__.to_s  # XNvg̕R[h
      c = ecnv[etype.downcase]
      str = NKF.nkf("-m0#{c}", str)  if c != nil
      str.force_encoding(etype)
    end
    return str
  end

  def time2str(x)
    if x.class == Time
      x = x.to_s.sub(/ +[\+\-]\d+$/, "")
      x = x.sub(/( +|\:)0+$/, "") while x =~ /( +|\:)0+$/
      x = x + ":00"  if x =~ / +\d+$/
    elsif x.class == Array
      for i in 0...x.size
        x[i] = time2str(x[i])
      end
    elsif x.class == Hash
      hs = {}
      x.each {|key, val|
        hs[time2str(key)] = time2str(val)
      }
      x.clear
      x = hs
    end
    return x
  end

  def to_num(x)
    if x.class == String
      if x =~ /^-?\d+$/
        x = x.to_i
      elsif x =~ /^(-?\d+)\.0+$/
        x = $1.to_i
      elsif x =~ /^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
        x = x.to_f
      elsif x =~ /^\d{4}[\/\-]\d\d?[\/\-]\d\d?( +\d\d?:\d\d?(:\d\d?)?)?$/ or
        x =~ /^\d{4}[\/\-]\d\d?[\/\-]\d\d? +\d\d?:\d\d?:\d\d?( +[\+\-]?\d+)?$/
        x = Time.parse(x)
      end
    elsif x.class == Array
      for i in 0...x.size
        x[i] = to_num(x[i])
      end
    elsif x.class == Hash
      hs = {}
      x.each {|key, val|
        hs[to_num(key)] = to_num(val)
      }
      x.clear
      x = hs
    end
    return x
  end

  def str2ary(str, sep = nil, atype = :num)  # zɕϊ
    sep = RrxSub.get_sep(str)  if sep == nil
    ary = (RUBY_VERSION < "1.9") ? CSV.parse(str, sep) :
        CSV.parse(str, {:col_sep=>sep})
    ary.map! {|row|
      row.map! {|x|
        x = x.to_s
        if x =~ /^\s*$/
          x = nil
        end
        x
      }
      row.pop  while row.size>0 and row.last==nil
      row
    }
    ary = to_num(ary)  if atype == :num
    return ary
  end

  def ary2str(aa, sep=',')  # z𕶎ɕϊ
    aa = time2str(aa)
    if aa.class == Array and aa[0].class != Array
      aa = [aa]
    end
    aa.map! {|row|
      row.map! {|x|
        if x.to_s =~ /^\s*$/
          x = nil
        end
        x
      }
    }
    str = ''
    aa.each {|row|
      row = [row]  if row.class != Array
      line = (RUBY_VERSION < "1.9") ? CSV.generate_line(row, sep) :
          CSV.generate_line(row, {:col_sep=>sep})
      line.chomp!
      str += "#{line}\n"
    }
    return str
  end

  def temp_make(*arg)  # e|t@C쐬
    last_flag =  (arg.size>0 and [true,false].include?(arg.last)) ?
        arg.pop : true
    str = ext = nil
    arg.each {|x|
      if x.class == Symbol or (x.class == String and
          x =~ /^\.[a-z0-9]+$/i)
        if ext == nil
          ext = x.to_s
          ext = '.' + ext  unless ext =~ /^\./
        end
      elsif x.class == String or x.class == Array
        str = x  if str == nil
      end
    }
    if str.class == Array
      sep = (ext.to_s.downcase == '.txt') ? "\t" : ","
      str = ary2str(str, sep)
    end
    w = @@prefix
    w = [w, ext]  if ext
    ff = Tempfile.new(w)
    ff.write str  if str != nil
    ff.close
    @@temp_hs[ff.path] = ff
    @@file_hs['last'] = ff  if last_flag == true
    return ff
  end

  def temp_read(ff)  # e|t@C̓e𕶎ŕԂ
    ff.open
    str = ff.read
    ff.close
    return str
  end

  def temp_write(ff, str)  # e|t@Cɏ
    res = nil
    if ff and str
      ff.open
      res = ff.write str
      ff.close
    end
    return res
  end

  def temp_delete(ff = nil)  # e|t@C̍폜
    if ff != nil
      key = (RUBY_VERSION < "1.9") ? @@temp_hs.index(ff) : @@temp_hs.key(ff)
      @@temp_hs.delete(key)  if key and @@temp_hs[key]
      key = (RUBY_VERSION < "1.9") ? @@file_hs.index(ff) : @@file_hs.key(ff)
      @@file_hs.delete(key)  if key and @@file_hs[key]
      ff.close
      ff.unlink
      return
    end
    if @@file_hs.size > 0
      @@file_hs.each {|key, ff|
        ff.close
        ff.unlink
      }
      @@file_hs.clear
    end
    if @@temp_hs.size > 0
      @@temp_hs.each {|fname, ff|
        next  unless test(?e, fname)
        ff.close
        ff.unlink
      }
      @@temp_hs.clear
    end
  end

  def temp_delete_all()  # ֘Ae|t@C̑S폜
    temp_delete()
    ff = temp_make(false)
    dd = File.dirname(ff.path)
    temp_delete(ff)
    dd = dd.gsub(/\\/, "/")
    dd.sub(/\/$/, "")
    Dir.glob("#{dd}/#{@@prefix}*") {|fname|
      File.unlink(fname)
    }
  end

  def temp_file(key)  # keyɑΉTempfileIuWFNgԂ
    ff = nil
    if key
      ff = @@file_hs[key]
      unless ff
        ff = @@temp_hs[key]
      end
    end
    return ff
  end

  def temp_key()  # TempfilekeyQԂ
    res = []
    fa = []
    @@file_hs.each {|key, ff|
      res << key
      fa << ff
    }
    @@temp_hs.each {|key, ff|
      next  if fa.include?(ff)
      res << key
    }
    return res
  end

  def temp_keyset(ff, key)
    @@file_hs[key] = ff
  end

  def rsave(hs = nil)  # rrx.cfg̕ۑ
    if hs.class == Hash and hs.size > 0
      str = ''
      hs.each {|key, val|
        case key
        when :path
          str = str + "path\t#{val}\n"  if test(?e, val)
        when :opt
          str = str + "opt\t#{val}\n"
        when :log
          str = str + "log\t#{val}\n"
        when :encopt
          str = str + "encopt\t#{val}\n"
        end
      }
      if str != ''
        File.open(@@rrx_cfg, "w") {|ff|  ff.write str}
      end
    else
      File.open(@@rrx_cfg, "w") {|ff|
        ff.print "path\t#{@@path}\n" +
            "opt\t#{@@opt}\n" +
            "log\t#{@@log}\n" +
            "encopt\t#{@@encopt}\n"
      }
    end
  end

  def rload(sw = nil)  # rrx.cfg̓Ǎ
    hs = {}
    if !test(?f, @@rrx_cfg) or (str = file_read(@@rrx_cfg)) == ''
      return hs
    end
    str.split(/\n/).each {|line|
      line = line.sub(/^\s+/, "").sub(/\s+$/, "")
      next  unless line =~ /\t/
      a = line.split(/\t/)
      case a[0]
      when 'path'
        wpath = (a[1] =~ /^\s*$/) ? nil : a[1]
        hs[:path] = wpath  if wpath
      when 'opt'
        v = (a[1] =~ /^\s*$/) ? nil : a[1]
        hs[:opt] = v  if v
      when 'log'
        hs[:log] = (a[1] =~ /^\s*$/) ? nil : a[1]
      when 'encopt'
        if a[1] =~ /true/i
          hs[:encopt] = true
        elsif a[1] =~ /false/i
          hs[:encopt] = false
        else
          hs[:encopt] = nil
        end
      end
    }
    if sw == :set
      hs.each {|key, val|
        case key
        when :path  then  @@path=val if @@path == nil;
        when :opt  then  @@opt=val if @@opt == nil;
        when :log  then  @@log=val if @@log == nil;
        when :encopt  then  @@encopt=val if @@encopt == nil;
        end
      }
    end
    return hs
  end

  def search_rcmd_path()
    res = nil
    unless defined?(WIN32OLE)
      return res
    end
    if ENV['RCMD_PATH'] and test(?e, ENV['RCMD_PATH'])
      res = ENV['RCMD_PATH'].gsub(/\\/, "/")
      return res
    end
    whome = WIN32OLE.new("WScript.Shell").SpecialFolders("MyDocuments")
    whome = whome.sub(/^[A-Z]\:/i, "").gsub(/\\/, "/") + "/R"
    dir_list = ['/R', '/Program Files/R', '/Program Files (x86)/R', whome]
    file_system = WIN32OLE.new("Scripting.FileSystemObject")
    drives = file_system.Drives
    drives.each do |drive|
      next  unless drive.IsReady
      dl = "#{drive.DriveLetter}"
      base_dir = nil
      dir_list.each do |dd|
        w = "#{dl}:#{dd}"
        if test(?d, w)
          base_dir = w
          break
        end
      end
      next  unless base_dir
      ary = []
      Dir.glob("#{base_dir}/R-*") {|dname|
        next  unless test(?d, dname)
        fname = "#{dname}/bin/i386/Rcmd.exe"
        if test(?e, fname)
          ary << fname
        end
      }
      if ary.size > 0
        if ary.size > 1
          nl = '000.000'
          vl = nil
          ary.each {|val|
            wval = val.sub(/^.+\/R-(\d.+?)\/bin\/i386\/Rcmd\.exe$/i, "\\1")
            nary = wval.split(/[^0-9]+/)
            nary.map! {|n|
              n = sprintf("%03d", n.to_i)
            }
            ns = nary.join(".")
            if ns > nl
              nl = ns
              vl = val
            end
          }
          res = vl
        else
          res = ary[0]
        end
        break
      end
    end
    return res
  end

  def delmargin(ary)  # ̍v폜
    aa = []
    ary.each {|row|  aa.push(row.dup)}
    mr = aa.size-1
    mc = aa[0].size-1
    if aa[0][0].to_s =~ /^-?\d+$|^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
      rs=0;  cs=0
    else
      rs=1;  cs=1
    end
    flag = true
    for i in rs..mr
      unless (aa[i].size-1) == mc
        flag = false
        break
      end
      ttl = 0.0
      for j in cs...mc  # E[uvv𒲂ׂ
        ttl += aa[i][j].to_f
      end
      unless ttl == aa[i][mc].to_f
        flag = false
        break
      end
    end
    if flag
      for i in 0..mr
        aa[i].delete_at(mc)
      end
      return aa
    end
    flag = true
    for i in rs..mr
      unless (aa[i].size-1) == mc
        flag = false
        break
      end
      ttl = 0.0
      for j in (cs+1)..mc  # [uvv𒲂ׂ
        ttl += aa[i][j].to_f
      end
      unless ttl == aa[i][cs].to_f
        flag = false
        break
      end
    end
    if flag
      for i in 0..mr
        aa[i].delete_at(cs)
      end
    end
    return aa
  end

  def del_sum(aa)  # cE̍v폜
    flag = :array
    sep = nil
    if aa.class == String
      sep = RrxSub.get_sep(aa)
      aa = str2ary(aa, sep)
      flag = :string
    end
    aa = delmargin(aa)
    aa = RrxSub.table_turn(aa)
    aa = delmargin(aa)
    aa = RrxSub.table_turn(aa)
    if flag == :string
      res = ary2str(res, sep)
    end
    return aa
  end

  def cut_num(*args, &block)  # s(l̒c)𒊏oAv폜
    res = []
    tfiles = []
    while arg = args.shift
      if arg.class == String and arg !~ /\r|\n/ and
          arg =~ /^http.?\:\/\/|^ftp\:\/\//i
        w = get_web(arg)
        if w != nil
          args.unshift(w.path)
          tfiles << w
        else
          STDERR.puts "can't read uri: '#{arg}'"
        end
        next
      end
      if arg.class == String and arg !~ /\r|\n/ and arg =~ @@ext_regex
        ww = xls2ary(arg)
        if ww.size > 0
          while w = ww.pop
            args.unshift(w)
          end
        else
          STDERR.puts "can't open or read the file '#{arg}'"
        end
        next
      end
      flag = nil
      if arg.class == String and arg !~ /\r|\n/
        if  !test(?f, arg) or (body = file_read(arg)) == ''
          STDERR.puts "can't open the file '#{arg}'"
          next
        else
          flag = :file
        end
      else
        body = arg
      end
      if body.class == String
        sep = RrxSub.get_sep(body)
        ary = str2ary(body)
        flag = :string
      elsif body.class == Array
        ary = body
        flag = :array
      end
      if flag == nil
        next
      end
      arys = RrxSub.select_squares(ary, :num)
      arys.each {|wary|
        w = del_sum(wary)
        if flag == :string or flag == :file
          w = ary2str(w, sep)
        end
        block.call(w)  if block
        res << w
      }
    end
    tfiles.each {|tff|  tff.unlink}  if tfiles.size > 0
    return res
  end

  def cut_num2(*args, &block)  # NXWv\𒊏oAv폜
    res = []
    rcv = 1
    if [1,2,3].include?(args.last)
      rcv = args.pop
    end
    tfiles = []
    while arg = args.shift
      if arg.class == String and arg !~ /\r|\n/ and
          arg =~ /^http.?\:\/\/|^ftp\:\/\//i
        w = get_web(arg)
        if w != nil
          args.unshift(w.path)
          tfiles << w
        else
          STDERR.puts "can't read uri: '#{arg}'"
        end
        next
      end
      if arg.class == String and arg !~ /\r|\n/ and arg =~ @@ext_regex
        ww = xls2ary(arg)
        if ww.size > 0
          while w = ww.pop
            args.unshift(w)
          end
        else
          STDERR.puts "can't open or read the file '#{arg}'"
        end
        next
      end
      flag = nil
      if arg.class == String and arg !~ /\r|\n/
        if  !test(?f, arg) or (body = file_read(arg)) == ''
          STDERR.puts "can't open the file '#{arg}'"
          next
        else
          flag = :file
        end
      else
        body = arg
      end
      if body.class == String
        sep = RrxSub.get_sep(body)
        ary = str2ary(body)
        flag = :string
      elsif body.class == Array
        ary = body
        flag = :array
      end
      if flag == nil
        next
      end
      arys = RrxSub.select_num2(ary)
      arys.each {|wary, title, read_opt|
        w = del_sum(wary)
        if flag == :string or flag == :file
          w = ary2str(w, sep)
        end
        if rcv == 1
          block.call(w)  if block
          res << w
        elsif rcv == 2
          block.call(w, title)  if block
          res << [w, title]
        elsif rcv == 3
          block.call(w, title, read_opt)  if block
          res << [w, title, read_opt]
        end
      }
    end
    tfiles.each {|tff|  tff.unlink}  if tfiles.size > 0
    return res
  end

  def cut_data(*args, &block)  # s̒o
    res = []
    rcv = 1
    if [1,2,3].include?(args.last)
      rcv = args.pop
    end
    tfiles = []
    while arg = args.shift
      if arg.class == String and arg !~ /\r|\n/ and
          arg =~ /^http.?\:\/\/|^ftp\:\/\//i
        w = get_web(arg)
        if w != nil
          args.unshift(w.path)
          tfiles << w
        else
          STDERR.puts "can't read uri: '#{arg}'"
        end
        next
      end
      if arg.class == String and arg !~ /\r|\n/ and arg =~ @@ext_regex
        ww = xls2ary(arg)
        if ww.size > 0
          while w = ww.pop
            args.unshift(w)
          end
        else
          STDERR.puts "can't open or read the file '#{arg}'"
        end
        next
      end
      flag = nil
      if arg.class == String and arg !~ /\r|\n/
        if  !test(?f, arg) or (body = file_read(arg)) == ''
          STDERR.puts "can't open the file '#{arg}'"
          next
        else
          flag = :file
        end
      else
        body = arg
      end
      if body.class == String
        sep = RrxSub.get_sep(body)
        ary = str2ary(body)
        flag = :string
      elsif body.class == Array
        ary = body
        flag = :array
      end
      if flag == nil
        next
      end
      arys = RrxSub.select_data2(ary)
      arys.each {|w, title, read_opt|
        if flag == :string or flag == :file
          w = ary2str(w, sep)
        end
        if block
          if rcv == 1
            block.call(w)
            res << w
          elsif rcv == 2
            block.call(w, title)
            res << [w, title]
          elsif rcv == 3
            block.call(w, title, read_opt)
            res << [w, title, read_opt]
          end
        end
      }
    end
    tfiles.each {|tff|  tff.unlink}  if tfiles.size > 0
    return res
  end

  def select_cells(ary, condition, flag=nil, &block)  # YZ̒o
    res = Array.new
    col = ary[0]
    for i in 1...ary.size
      row = ary[i]
      for j in 1...row.size
        value = row[j]
        if eval(condition)
          if flag == :addr or flag == :address
            res << [row[0], col[j], [i,j]]
          else
            res << [row[0], col[j], value]
          end
          block.call(res.last)  if block
        end
      end
    end
    return res
  end

  def table_turn(aa)
    RrxSub.table_turn(aa)
  end

  def del_row(ary, *nums)  # Qz񒆂̎ws̍폜
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    item_w = RrxSub.table_turn(ary)[0]
    item_sz = item_w.size
    na = []
    nums.map! {|x|
      if x.class == String
        item_w.each_with_index {|s, n|
          if s == x
            na << n  unless nums.include?(n)
          end
        }
        x = nil
      elsif x.class == Regexp
        item_w.each_with_index {|s, n|
          if s =~ x
            na << n  unless nums.include?(n)
          end
        }
        x = nil
      elsif x.class == Fixnum and x < 0
        x = item_sz + x
      end
      x
    }
    nums.delete(nil)
    nums = nums + na
    res = []
    ary.each_with_index {|row, i|
      next  if nums.include?(i)
      res << row.dup
    }
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def del_col(ary, *nums)  # Qz񒆂̎w̍폜
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    item_w = ary[0]
    item_sz = item_w.size
    na = []
    nums.map! {|x|
      if x.class == String
        item_w.each_with_index {|s, n|
          if s == x
            na << n    unless nums.include?(n)
          end
        }
        x = nil
      elsif x.class == Regexp
        item_w.each_with_index {|s, n|
          if s =~ x
            na << n  unless nums.include?(n)
          end
        }
        x = nil
      elsif x.class == Fixnum and x < 0
        x = item_sz + x
      end
      x
    }
    nums.delete(nil)
    nums = nums + na
    aa = RrxSub.table_turn(ary)
    res = []
    aa.each_with_index {|col, i|
      next  if nums.include?(i)
      res << col
    }
    res = RrxSub.table_turn(res)
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def ins_row(ary, n, *arg)
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    if n.class == Fixnum and n < 0
      n = ary.size + n
    end
    ins_sw = false
    res = []
    ary.each_with_index {|row, i|
      if i == n
        arg.each {|x|  res << x.dup}
        ins_sw = true
      end
      res << row.dup
    }
    unless ins_sw
      arg.each {|x|  res << x.dup}
    end
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def ins_col(ary, n, *arg)
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    aa = RrxSub.table_turn(ary)
    if n.class == Fixnum and n < 0
      n = aa.size + n
    end
    ins_sw = false
    res = []
    aa.each_with_index {|row, i|
      if i == n
        arg.each {|x|  res << x.dup}
        ins_sw = true
      end
      res << row
    }
    unless ins_sw
      arg.each {|x|  res << x.dup}
    end
    res = RrxSub.table_turn(res)
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def clp2str()  # Nbv{[h̓e𕶎ƂĎ擾
    str = nil
    if defined?(WIN32OLE)
      str = rencode( Rclip::getData.to_s )
      str = str.gsub(/\r\n/, "\n")
    end
    return str
  end

  def str2clp(str)  # Nbv{[hɏo
    if defined?(WIN32OLE)
      Rclip::setData(str)
    end
  end

  def xls2ary(target, fmt = :array)
    res = nil
    tfile = nil
    if target =~ /^http.?\:\/\/|^ftp\:\/\//i
      w = get_web(target)
      if w != nil
        tfile = w
        target = tfile.path
      end
    end
    if !test(?e, target)
      STDERR.puts "can't find '#{target}'"
      return []
    end
    ext = File.extname(target).downcase
    err_msg = nil
    if defined?(Spreadsheet) and ext == ".xls"
      begin
        book = Spreadsheet.open(target, "rb")
      rescue
        err_msg = "Error in Spreadsheet.open() " +
            "<< xls2ary():\n" + $! + "\n"
        book = nil
      end
      if book != nil
        res = (fmt == :hash) ? {} : []
        book.worksheets.each do |ss|
          ary = []
          wlng = []
          ss.each do |row|
            row2 = []
            row.each do |cell|
              row2 << cell
            end
            ary << row2
            if row2.size > 0
              n = 0
              while row2[n] == nil and n < row2.size
                n += 1
              end
              wlng << n
            end
          end
          next  if ary == []
          ary = str2ary( rencode( ary2str(ary) ) )
          if wlng.size > 0 and !wlng.include?(0)
            min = wlng.shift
            while n = wlng.shift
              min = n  if n < min
            end
            ary.map! {|row|
              row = (!row or row.size < 1) ? [] : row[min..-1]
            }
          end
          if fmt == :hash then  res[rencode(ss.name)] = ary
          else  res << ary;  end
        end
      end
    end
    if res == nil and defined?(Roo) and
        %w(.xls .xlsx .xml .ods).include?(ext)
      ext_sym = ext.sub(/^\./, "").to_sym
      begin
        book = Roo::Spreadsheet.open(target, :extension=>ext_sym)
      rescue
        err_msg = "Error in Roo::Spreadsheet.open() " +
            "<< xls2ary():\n" + $! + "\n"
        book = nil
      end
      if book != nil
        res = (fmt == :hash) ? {} : []
        book.sheets.each do |ss|
          book.default_sheet = ss
          next  unless book.first_row
          ary = []
          for row in book.first_row..book.last_row
            row2 = []
            for col in book.first_column..book.last_column
              row2 << book.cell(row,col)
            end
            ary << row2
          end
          ary = str2ary( rencode( ary2str(ary) ) )
          if fmt == :hash then  res[rencode(ss)] = ary
          else  res << ary;  end
        end
      end
    end
    if res == nil and defined?(Exlap)
      xl = Exlap.new
      if xl.obj != nil and xl.ext_type[ext]
        res = (fmt == :hash) ? {} : []
        xl.opens(target) do |wb|
          wb.each do |ss|
            next  if ss.empty?
            ary = ss.range_to_a
            ary = str2ary( rencode( ary2str(ary) ) )
            if fmt == :hash then  res[rencode(ss.Name)] = ary
            else  res << ary;  end
          end
        end
        xl.quit
      end
    end
    if res == nil
      unless err_msg
        err_msg = "Error in xls2ary(): must require " +
            "'spreadsheet' | 'roo' | 'exlap'\n"
      end
      STDERR.print err_msg
      return res
    end
    tfile.unlink  if tfile != nil
    res = nil  if res.size == 0
    if res == nil
      unless err_msg
        err_msg = "Error in xls2ary(): can't read '#{target}'\n"
      end
      STDERR.print err_msg
    end
    return res
  end

  def gss2ary(target, opt=nil, fmt = :array)
    res = nil
    err_msg = nil
    begin
      book = !opt ? Roo::Google.new(target) : Roo::Google.new(target, opt)
    rescue
      err_msg = "Error in Roo::Google.new() " +
          "<< gss2ary():\n" + $! + "\n"
      book = nil
    end
    if book != nil
      res = (fmt == :hash) ? {} : []
      book.sheets.each do |ss|
        book.default_sheet = ss
        next  unless book.first_row
        ary = []
        for row in book.first_row..book.last_row
          row2 = []
          for col in book.first_column..book.last_column
            row2 << book.cell(row,col)
          end
          ary << row2
        end
        ary = str2ary( rencode( ary2str(ary) ) )
        if fmt == :hash then  res[rencode(ss)] = ary
        else  res << ary;  end
      end
      res = nil  if res.size == 0
    end
    if res == nil
      unless err_msg
        err_msg = "Error in gss2ary: can't read " +
            "'#{target}' by Roo::Google.new()\n"
      end
      STDERR.print err_msg
    end
    return res
  end

  def robj_convert_sub(str)
    res = {}
    str.split(RS).each do |xx|
      a = xx.split(FS)
      next  if a.class != Array or a[0] == nil
      if a[1] == CSV_MARK
        key, dummy, val = a
        res[key] = time2str(str2ary(val, ","))
      else
        key, val = a
        if val
          if val.include?(VFS)
            val = val.split(VFS)
            val.delete("")
            val.delete(nil)
            val = time2str(to_num(val))
          else
            val = time2str(to_num(val))
          end
        end
        res[key] = val
      end
    end
    return res
  end

  def robj_convert(str)
    regex = Regexp.new("#{USS}([^#{USS}#{UTT}]+)#{UTT}")
    ss = "__!_!!__"
    tt = "__#_##__"
    count = -1
    ary = []
    while str =~ regex
      w = $1
      count += 1
      ary[count] = robj_convert_sub(w).inspect
      str = str.sub(regex, "#{ss}#{count}#{tt}")
    end
    str = str.inspect
    while str =~ /#{ss}(\d+)#{tt}/
      n = $1.to_i
      str = str.sub(/#{ss}#{n}#{tt}/, "#{USS}#{ary[n]}#{UTT}")
    end
    while str =~ /"#{regex}"/
      w = $1
      w = w.gsub(/#{USS}|#{UTT}/, "")
      str = str.sub(/"#{regex}"/, w)
    end
    while str =~ /"(#{regex}{2,})"/
      w = $1
      w = w.gsub(/\}#{UTT}#{USS}\{/, ", ")
      w = w.gsub(/#{USS}|#{UTT}/, "")
      str = str.sub(/"(#{regex}{2,})"/, "#{w}")
    end
    res = nil
    str = "res = " + str
    eval(str)
    return res
  end

  def charset(smp = nil)
    res = nil
    if RUBY_VERSION < "1.9"
      c = $KCODE.to_s.downcase[0,1]
      hs = {'e'=>'euc-jp', 's'=>'cp932', 'u'=>'utf-8'}
      res = hs[c]
    else
      res = !smp ? __ENCODING__.to_s : smp.encoding.to_s
    end
    return res
  end

  def r_each(argsize=2, &block)
    argsize = 2  if argsize != 2 and argsize != 3
    hs = @@r_result
    if !block or hs.class != Hash or hs.size < 1
      return
    end
    @@r_keys.each do |name|
      xx = hs[name]
      case argsize
      when 2
        block.call(name, xx)
      when 3
        if xx.class == Hash
          xx.each {|key, val|  block.call(name, key, val)}
        else
          block.call(name, nil, xx)
        end
      end
    end
  end

  def robj_each(argsize=2, &block)
    argsize = 2  if argsize != 2 and argsize != 3
    hs = @@r_result
    if !block or hs.class != Hash or hs.size < 1
      return
    end
    @@robj_list.each do |mtm, key0, name|
      xx = hs[name]
      case argsize
      when 2
        block.call(name, xx)
      when 3
        if xx.class == Hash
          xx.each {|key, val|  block.call(name, key, val)}
        else
          block.call(name, nil, xx)
        end
      end
    end
  end

  def get_web(url, res_type = nil, opt = nil)
    uri = URI.parse(url)
    os_type = (uri.scheme != "http") ? :COMMON : nil
    ext = File.extname(uri.path)
    ext = nil  if ext.to_s == ''
    res_type = :file  unless res_type
    opt = @@http_opt  unless opt
    mshttp = RrxSub.mshttp_obj(os_type)
    if mshttp == nil
      begin
        sio = !opt ? OpenURI.open_uri(url) : OpenURI.open_uri(url, opt)
        body = sio.read
      rescue
        STDERR.puts "Error in get_web(): #{$!}"
        body = nil
      end
    else  # if mshttp != nil
      sec_limit = 50  # 0.2*50=10 seconds
      if url =~ /^http/i
        mshttp.Open "GET", url, false
        sec_count = 0
        while mshttp.readyState != 1
          sleep 0.2
          sec_count += 1
          break  if sec_count > sec_limit
        end
        if sec_count > sec_limit
          mshttp.abort
          return nil
        end
        begin
          mshttp.Send
        rescue
          mshttp.abort
          return nil
        end
        sec_count = 0
        while mshttp.readyState != 4
          sleep 0.2
          sec_count += 1
          break  if sec_count > sec_limit
        end
        if sec_count > sec_limit
          mshttp.abort
          return nil
        end
        body = (mshttp.status == 200) ? mshttp.responseBody.pack("c*") : nil
        mshttp.abort
      end
    end
    res = body
    if res != nil and res_type == :file
      tff = temp_make(ext, false)
      File.open(tff.path, "wb") {|ff|  ff.write body}
      res = tff
    end
    return res
  end

  def http_opt(opt)
    @@http_opt = opt
  end

  def to_rba(xx, type)  # Rarray|matrixruby̔z`ɕϊ
    if ![:array, :matrix].include?(type) or
        xx.class != Array or xx[0].class != Array
      return xx
    end
    top_left = xx[0][0]
    res = xx
    if res.size > 1 and res[0].size > 1 and top_left == nil
      n_flag = true
      for i in 1...res.size
        if res[i][0] != i
          n_flag = false
          break
        end
      end
      if n_flag
        for i in 0...res.size
          res[i] = res[i][1..-1]
        end
      end
    end
    if type == :array  # R̂Pz̏ꍇ
      res.delete_at(0)  if top_left == nil
      res = RrxSub.table_turn(res)
    end
    return res
  end

  def to_rex(x)  # rubyString, ArrayR̕EvectorEmatrixɕϊ
    res = nil
    if x.class == String
      res = x.inspect
    elsif x.class == Array and x[0].class != Array  # Pz
      w = x.inspect
      res = "c(" + w[1..-2] + ")"
    elsif x.class == Array and x[0].class == Array and  # Qz
          x[0][0] != Array
      ncol = 0
      x.each {|row|
        n = row.size
        ncol = n  if n > ncol
      }
      x2 = []
      x.each {|row|
        x2 << (row + [""]*ncol)[0..(ncol-1)]
      }
      m = ''
      x2.each {|v|
        w = v.inspect
        m = m + w[1..-2] + ", "
      }
      v = "c(" + m[0..-3] + ")"
      res = "matrix(" + v + ", ncol=#{ncol}, byrow=T)"
    end
    return res
  end

  def nkfex(str, enc="utf-8")  # string encode by nkf
    enc = enc.to_s.downcase
    cc = ''
    case enc
    when 'euc-jp' then cc = 'e'
    when 'jis', 'iso-2022-jp' then cc = 'j'
    when 'shift_jis', 'sjis', 'windows-31j', 'cp932' then cc = 's'
    when 'utf-8', 'utf8' then cc = 'w'
    when 'utf-16', 'utf16' then cc = 'w16'
    else
      i = %w(e s u).index(enc[0,1])
      if i != nil
        cc = %w(e s w)[i]
      end
    end
    if cc != ''
      str = NKF.nkf("-m0#{cc}", str)
      if RUBY_VERSION >= "1.9" and cc == 's' and
          %w(windows-31j cp932).include?(enc)
        str.force_encoding("Windows-31J")
      end
    end
    return str
  end

  def web2ary(url, reg=nil, count=nil)
    return Rrxweb.web2ary(url, reg, count)
  end

  def VERSION()
    puts "1.1"
  end

  def xls2tmp(target, sn=1, ftype=".csv")
    sn = sn - 1
    sn = 0  if sn < 0
    arys = xls2ary(target)
    ary = (arys.class == Array) ? arys[sn] : nil
    tfile = nil
    if ary.class == Array and ary[0].class == Array
      ary.shift  while ary[0].size < 2
      ary.pop  while ary[-1].size < 2
      tfile = temp_make(ary, ftype, false)  if ary.size > 1
    end
    return tfile
  end

  def htm2tmp(target, reg=nil, sn=1, ftype=".csv")
    sn = sn - 1
    sn = 0  if sn < 0
    arys = web2ary(target, reg, 2)
    ary = (arys.class == Array) ? arys[sn] : nil
    tfile = nil
    if ary.class == Array and ary[0].class == Array
      ary.shift  while ary[0].size < 2
      ary.pop  while ary[-1].size < 2
      tfile = temp_make(ary, ftype, false)  if ary.size > 1
    end
    return tfile
  end

  def web2file(url)
    body = get_web(url, :string)
    fname = nil
    if body.class == String and body != ''
      uri = URI.parse(url)
      fname = File.basename(uri.path)
      if fname == nil or fname == ''
        fname = "index.htm"
      else
        fname = fname.gsub(/[\\\/\|:"'`<>]/, "_")
      end
      File.open(fname, "wb") {|ff|  ff.write body}
    end
    return fname
  end

  def fullpath(fname, sep = "\\")  # w̃t@CtpXŕԂ
    if defined?(WIN32OLE)
      fso = WIN32OLE.new('Scripting.FileSystemObject')
      w = fso.GetAbsolutePathName(fname)
      if sep != "\\"
        w = w.gsub(/\\/, sep)
      end
    else
      w = fname
    end
    return w
  end

  def fullpath2(fname)
    return fullpath(fname, "/")
  end

  alias :s2a :str2ary
  alias :a2s :ary2str
  alias :c2s :clp2str
  alias :s2c :str2clp
  alias :cut_table :cut_num2
  alias :del_yoko :del_row
  alias :del_tate :del_col
  alias :del_horizontal :del_row
  alias :del_vertical :del_col
  alias :ins_yoko :ins_row
  alias :ins_tate :ins_col
  alias :ins_horizontal :ins_row
  alias :ins_vertical :ins_col
  alias :x2a :xls2ary
  module_function(:s2a, :a2s, :c2s, :s2c, :x2a, :cut_table)
  module_function(:del_yoko, :del_tate, :del_horizontal, :del_vertical)
  module_function(:ins_yoko, :ins_tate, :ins_horizontal, :ins_vertical)
end


# methods for web
module Rrxweb
  def self.init()
    enc_org = (RUBY_VERSION < "1.9") ? $KCODE : __ENCODING__
    enc = enc_org.to_s.downcase
    cc = ''
    case enc
    when 'euc-jp' then cc = 'e'
    when 'jis', 'iso-2022-jp' then cc = 'j'
    when 'shift_jis', 'sjis', 'windows-31j', 'cp932' then cc = 's'
    when 'utf-8', 'utf8' then cc = 'w'
    when 'utf-16', 'utf16' then cc = 'w16'
    else
      i = %w(e s u).index(enc[0,1])
      if i != nil
        cc = %w(e s w)[i]
      end
    end
    nkf_c = (cc != '') ? cc : 'w'
    return [enc_org, nkf_c]
  end

  @enc, @nkfc = init()

  # ȈnkfĂяo
  def self.xnkf(str, code = @nkfc)
    res = NKF.nkf("-m0#{code}", str)
    if RUBY_VERSION >= "1.9" and code == 's' and
       @enc.to_s =~ /windows-31j|cp932/i
      res.force_encoding("Windows-31J")
    end
    return res
  end

  # lQ(&#cc;)𕶎ɕϊ
      # o͕R[h @nkfc (j, e, s, w Ȃ)ɏ]B
  def self.ncr_convert(str)
    unless str =~ /\&\#.+;/
      return str
    end
    str = str.gsub(/\&\#x([0-9a-f]+);/i) {|ch|
      n = $1.hex
      ch = "&##{n};"
    }
    str.gsub!(/\&\#(\d+);/) {|ptn|
      n = $1.to_i
      if n > 0xff
        ch = NKF.nkf("-m0W16#{@nkfc}", [n].pack("n"))
      else
        ch = ptn
      end
      ptn = ch
    }
    return str
  end

  # htmltableruby̔zɕϊBZɑΉ
      # data_type(z̊evf̃^Cv)  :TEXT | :HTML
      # :HTML̏ꍇ <td>cc</td> z̊evfɑ
  def self.tbl2ary(tbl, data_type=:TEXT)
    aa = []
    if !defined?(Hpricot) and !defined?(Nokogiri)
      STDERR.puts "can't call \"Rrxweb::tbl2ary\" without " +
          "'hpricot' or 'nokogiri'"
      return aa
    end
    if (e = tbl.at("caption"))
      w = e.inner_text.sub(/^\s+/, "").sub(/\s+$/, "")
      w = "caption:" + w
      aa.push([w])
    end
    trs = tbl.search("tr")
    span = {}
    trs.each_with_index {|e, row_n|
      a = []
      col_n = 0
      e.search("*").each {|e2|
        next  unless e2.name =~ /^(th|td)$/i
        while span[[row_n, col_n]]
          a.push(nil)  # Žӏ̖ߑnilɂ
          col_n += 1
        end
        data = (data_type == :TEXT) ? e2.inner_text : e2.to_html
        data = data.sub(/\A\s+/, "")
        data = data.sub(/\s+\Z/, "")
        data = data.gsub(/( *\n *)+/, " | ")  if data_type == :TEXT
        cspan = 1
        if e2['colspan'] =~ /(\d+)/
          cspan = $1.to_i
        end
        rspan = 1
        if e2['rowspan'] =~ /(\d+)/
          rspan = $1.to_i
        end
        if cspan == 1 and rspan == 1  # ZƂ͖֌W
          a.push(data)
        else
          a.push(data)
          for r in row_n...(row_n+rspan)
            for c in col_n...(col_n+cspan)
              span[[r, c]] = true
            end
          end
        end
        col_n += 1
      }
      aa.push(a)
    }
    return aa
  end

  # K\̕R[hύX
  def self.arrange_regex(re, code)
    if re
      str = re.source
      opt = re.options
      str = xnkf(str, code)
      re = Regexp.new(str, opt)
    end
    return re
  end

  # w肳ꂽurlweby[W擾Ahtmlp[TpɏC
    # url '<' Ŏn܂ꍇ html source ̕Ƃ݂Ȃ
  def self.get_html(url)
    temp_chr = '__jfZR6XRr__'
    if url !~ /\r|\n/ and test(?f, url)
      url = File.open(url) {|ff| ff.read}
      url = xnkf(url)
    end
    body = ''
    if url =~ /^\s*</ and url =~ /<html/i and url =~ /<\/html>/i
      body = url.sub(/^\s+/, "")
    elsif url !~ /\r|\n/
      body = Rrx::get_web(url, :text)
    end
    if !body or body == ''
      STDERR.printf("could not access: %s\n", url)
      return ''
    end
    body = xnkf(body.gsub(/\r\n/, "\n"))
    body = ncr_convert(body)
    body.gsub!(/<pre[^>]*>.+?<\/pre>/im) {|mp|
      mp = mp.gsub(/\n/, "#{temp_chr}")
    }
    body.gsub!(/\n+/, "")
    re = /<(p|br|li|[oud]l|dt|dd|h[1-6]|div|center|blockquote|pre)([ >])/i
    body.gsub!(re, "\n<\\1\\2")
    body.gsub!(/<\/(p|li|[oud]l|dt|dd|h[1-6]|div|center|blockquote|pre)>/i,
        "</\\1>\n")
    body.gsub!(/<(hr|hr +[^>]+)>/i, "\n<\\1>\n")
    body.gsub!(/\n\n+/, "\n")
    body.gsub!(/#{temp_chr}/, "\n")
    body.gsub!(/<img[^>]+>/i) {|mp|  # 摜alt߂ɒu
      if mp =~ /alt\s*=\s*["'](.+?)["']/im
        w = $1
        w.sub!(/^\s+/, "")
        w.sub!(/\s+$/, "")
        mp = w
      end
    }
    body.gsub!(/\&nbsp;/i, "&#32;")
    return body
  end

  # wwebtable(ɍ)oĕԂ
  def self.web2ary(url, table_select_regexp=nil,
      table_select_count=nil)
    res_ary = Array.new
    if !defined?(Hpricot) and !defined?(Nokogiri)
      STDERR.puts "can't call \"Rrxweb::web2ary\" " +
          "without 'hpricot' or 'nokogiri'"
      return nil
    end
    body = get_html(url)
    if body == ''
      return nil
    end
    table_select_regexp = arrange_regex(table_select_regexp, @nkfc)
    kcode_org = $KCODE  if RUBY_VERSION < "1.9"
    change_flag = nil
    if defined?(Hpricot)
      doc = Hpricot(body)
    elsif defined?(Nokogiri)
      $KCODE = 'utf8'  if RUBY_VERSION < "1.9"
      unless @nkfc == 'w'
        body = xnkf(body, 'w')
        table_select_regexp = arrange_regex(table_select_regexp, 'w')
        change_flag = true
      end
      doc = Nokogiri::HTML.parse(body, nil, "UTF-8")
    end
    doc.search("table").each {|tbl_obj|
      next  if tbl_obj.at("table")  # tabletable͑ΏۊO
      tbl_str = tbl_obj.inner_html  # tablehtml source
      aa = tbl2ary(tbl_obj)
      next  if aa.size < 1
      if table_select_regexp
        unless table_select_regexp =~ tbl_str
          next
        end
      end
      if table_select_count
        if aa.size < table_select_count
          next
        end
      end
      res_ary << aa
    }
    $KCODE = kcode_org  if RUBY_VERSION < "1.9"
    if change_flag
      res_ary.map! {|aa|
        ss = Rrx::ary2str(aa)
        ss = xnkf(ss)
        aa = Rrx::str2ary(ss)
      }
    end
    return res_ary
  end
end


# append some methods to Rrx
module Rrx
    @@html_str = Hash.new
    @@html_str[:mathjax] = <<-RRX_MATHJAX
    <script src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML" type="text/javascript"></script>
    <meta http-equiv="X-UA-Compatible" content="IE=EmulateIE7" />
    RRX_MATHJAX
    @@html_str[:table] = <<-RRX_TABLE
    <style type="text/css">
    table, td, th { border: 2px #ff0000 solid; }
    </style>
    RRX_TABLE

  def html_string(param)
    str = @@html_str[param]
    str = str.gsub(/\n    /, "\n")
    return str
  end

  def md_table(mm, *arg)  # R's array | matrix to table of markdown
    mm_org = mm
    sep = opt = fsep = type = nil
    if arg.size > 0
      while w = arg.shift
        if (w.class == String and w =~ /^=+$/) or
            (w.class == Array and w.size == 1 and w[0] =~ /^=+$/)
          fsep = ["====="]
          next
        end
        sep = w  if w.class == Array
        opt = w  if w.class == String
        type = w  if w.class == Symbol
      end
    end
    if mm.class == String
      if mm !~ /\n/ and test(?f, mm)
        mm = str2ary(File.read(mm), nil, nil)
      else
        mm = str2ary(mm, nil, nil)
      end
    end
    mm = time2str(mm)
    mm = to_rba(mm, type)
    if mm.size >= 3
      flag = :separater
      mm[1].each do |elm|
        flag = :no  unless elm =~ /^(-|:)+$/
      end
      if flag == :separater
        sep = mm[1]
        mm.delete_at(1)
      end
    end
    if mm.class != Array or mm[0].class != Array or mm.size < 2
      STDERR.puts "invalid parameter '#{mm_org.inspect}' in Rrx::md_table()"
      return
    end
    sep = tbl_align(mm)  unless sep
    mm2 = (!sep or sep == []) ? mm : [mm[0], sep] + mm[1..-1]
    if fsep != nil and mm2[-2] != sep
      mm2 = mm2[0..-2] + [fsep, mm2[-1]]
    end
    res = ""
    mm2.each {|row|
      w = row.join("\x7f\x7f")
      w = w.gsub(/\|/, "\\|")
      w = w.gsub(/\x7f\x7f/, "|")
      res += "|#{w}\n"
    }
    opt = opt.to_s.sub(/\s+$/, "")
    if opt != ''
      res += "{: #{opt}}\n"
    end
    return res
  end

  def array2table(aa, *arg)  # R's array to table of markdown
    return md_table(aa, :array, *arg)
  end

  def matrix2table(aa, *arg)  # R's matrix to table of markdown
    return md_table(aa, :matrix, *arg)
  end
  alias :pipe_table :matrix2table

  def html_template(enc=nil, *ins)  # create a html template
    $rrx_enc = (enc.class == String) ? enc : nil
    ins_str = ''
    ins.each do |elm|
      if elm.class == String
        ins_str += elm
      elsif elm.class == Symbol and @@html_str[elm]
        ins_str += self.html_string(elm)
      end
    end
    template_str = <<-EOS
		<% if $rrx_enc != nil
		     if RUBY_VERSION >= '1.9'
		       enc = @converter.root.options[:encoding]
		       enc = Encoding.default_external unless enc
		     end
		     cset = $rrx_enc
		   elsif RUBY_VERSION >= '1.9'
		     enc = @converter.root.options[:encoding]
		     enc = Encoding.default_external unless enc
		     cset = enc.to_s =~ /windows-31j|cp932|sjis|shift.jis/i ?
		         'Shift_JIS' : enc.to_s
		   else
		     enc = $KCODE.to_s[0,1].downcase
		     enc = 'w' unless ['e', 's'].include?(enc)
		     cset = {'e'=>'euc-jp', 's'=>'Shift_JIS', 'w'=>'utf-8'}[enc]
		   end
		%>\
		<!DOCTYPE html>
		<html>
		<head>
		<meta http-equiv="Content-type" content="text/html; charset=<%= cset %>" />
		<% extend ::Kramdown::Utils::Html
		title = ''
		h = @converter.root.children.find {|c| c.type == :header}
		if h
		  collector = lambda {|c| c.children.collect {|cc| cc.type == :text ?
		  escape_html(cc.value, :text) : collector.call(cc)}.join('')}
		  title = collector.call(h)
		  title = title.to_s.encode(enc) if RUBY_VERSION >= '1.9'
		end %>\
		<title><%= title %></title>
		<meta name="generator" content="kramdown <%= ::Kramdown::VERSION %>" />
		#{ins_str}</head>
		<body>
		<%= @body %>
		</body>
		</html>
    EOS
    template_str = template_str.gsub(/\t\t/, "")
    ff = Tempfile.new("tpl_")
    ff.write template_str
    ff.close
    return ff
  end

  def tex_template(dclass=nil)  # create a LaTeX template file(temporary)
    dclass = nil  if dclass.class != String
    dclass = !dclass ? '\documentclass{jarticle}' : dclass.chomp
    template_str = <<-EOS
		#{dclass}
		\\usepackage{listings}
		\\usepackage{hyperref}
		<% @converter.data[:packages].each {|pkg| %>\
		\\usepackage{<%= pkg %>}
		<% } %>\
		<% if @converter.data[:packages].include?('fancyvrb') %>\
		\\VerbatimFootnotes
		<% end %>\
		<% if @converter.data[:packages].include?('acronym') %>\
		<%   @converter.root.options[:abbrev_defs].each_pair do |k,v| %>\
		\\acrodef{<%= @converter.normalize_abbreviation_key(k) %>}[\
		<%= k %>]{<%= @converter.escape(v) %>}
		<%   end %>\
		<% end %>\
		\\hypersetup{colorlinks=true,urlcolor=blue}
		\\begin{document}
		<%= @body %>
		\\end{document}
    EOS
    template_str = template_str.gsub(/\t\t/, "")
        ff = Tempfile.new("tpl_")
    ff.write template_str
    ff.close
    return ff
  end

      # zeLXg̕\ɕϊ
  def text_table(ary, header = nil, sep = nil)
    if !defined?(Terminal::Table)
      STDERR.puts "'terminal-table' has not loaded"
      return nil
    end
    ary_org = ary
    if ary.class == String
      if ary !~ /\n/ and test(?f, ary)
        ary = str2ary(File.read(ary), nil, nil)
      else
        ary = str2ary(ary, nil, nil)
      end
    end
    ary.delete([])  if ary.class == Array
    ary = time2str(ary)
    if ary.class != Array or ary[0].class != Array
      STDERR.puts "invalid parameter '#{ary_org.inspect}' in Rrx::text_table()"
      return
    end
    if RUBY_VERSION >= "1.9"
      enc = nil
      ary.each do |row|
        row.each do |cell|
          if cell.class == String
            enc = cell.encoding
            break
          end
        end
        break  if enc
      end
      enc = __ENCODING__  unless enc
      enc = enc.to_s.downcase
    end
    ary.map! do |row|
      row.map! do |cell|
        cell = cell.to_s
        cell = cell.encode('utf-8')  if RUBY_VERSION >= "1.9"
        cell
      end
    end
    headings = header ? ary.shift : nil
    param = headings ? {:headings => headings} : {}
    table = Terminal::Table.new(param) do |t|
      while ary.size > 1
        t.add_row ary.shift
        t.add_separator  if sep
      end
      t.add_row ary.shift
    end
    str = table.to_s
    str = str.encode(enc)  if enc != nil and RUBY_VERSION >= "1.9"
    if header == '='
      lines = str.split(/\n/)
      for i in 1...lines.size
        if lines[i] =~ /^\+-.+-\+$/
          lines[i] = lines[i].gsub(/-/, "=")
          break
        end
      end
      str = lines.join("\n")
    end
    str += "\n"
    str.force_encoding(enc)  if RUBY_VERSION >= "1.9"
    return str
  end
  alias :ascii_table :text_table

  def tbl_align(ary)
    if ary.class != Array or ary.size < 2 or ary[0].class != Array
      return
    end
    regex1 = /^-?\d+$/
    regex2 = /^-?\d+\.(\d+)$/
    res = [nil] * ary[0].size
    for i in 1...ary.size
      for j in 0...ary[i].size
        next  if res[j] == :left
        cell = ary[i][j].to_s
        if cell =~ regex1
          case res[j]
          when nil, 0
            res[j] = 0
          else
            res[j] = :left
          end
        elsif cell =~ regex2
        n = $1.length
          case res[j]
          when nil, n
            res[j] = n
          else
            res[j] = :left
          end
        else
          res[j] = :left
        end
      end
    end
    for i in 0...res.size
      if res[i] == :left
        res[i] = "-----"
      else
        res[i] = "----:"
      end
    end
    return res
  end

  module_function(:html_string, :md_table, :array2table, :matrix2table,
      :pipe_table, :html_template, :tex_template, :text_table, :ascii_table,
      :tbl_align)
end

# append MarkDown methods to Rrx
module Rrx
  def kramdown_html(str, *opt)
    enc = (RUBY_VERSION >= "1.9") ? str.encoding() : nil
    hs = {}
    opt2 = []
    standalone_flag = false
    opt.each {|x|
      if x.class == Hash
        hs.update(x)
      elsif [:s, :standalone].include?(x)
        standalone_flag = true
      else
        opt2 << x
      end
    }
    opt = opt2
    tfile = nil
    if !hs[:template] and standalone_flag
      tfile = self.html_template(nil, *opt)
      hs[:template] = tfile.path
    end
    kdoc = Kramdown::Document.new(str, hs)
    html_str = kdoc.to_html
    html_str.force_encoding(enc)  if RUBY_VERSION >= "1.9" and enc != nil
    tfile.unlink  if tfile
    return html_str
  end

  def kramdown_tex(str, *opt)
    enc = (RUBY_VERSION >= "1.9") ? str.encoding() : nil
    hs = {}
    dclass = nil
    standalone_flag = false
    opt.each {|x|
      if x.class == Hash
        hs.update(x)
      elsif [:s, :standalone].include?(x)
        standalone_flag = true
      elsif x.class == String
        dclass = x
      end
    }
    tfile = nil
    if !hs[:template] and standalone_flag
      tfile = self.tex_template(dclass)
      hs[:template] = tfile.path
    end
    kdoc = Kramdown::Document.new(str, hs)
    tex_str = kdoc.to_latex
    tex_str.force_encoding(enc)  if RUBY_VERSION >= "1.9" and enc != nil
    tfile.unlink  if tfile
    return tex_str
  end

  def kramdown_contents(str)
    res = (str.to_s != '') ?
        "#{str}\n{:#contents}\n\n* contents list\n{:toc}\n" :
        "* contents list\n{:toc}\n"
    return res
  end

  def pandoc_convert(str, *opt_ary)
    unless defined?(PandocRuby)
      STDERR.puts "'pandoc-ruby' has not loaded."
      return nil
    end
    input_type = :markdown
    output_type = :html
    str = NKF.nkf("-m0xw", str)
    hs = {}
    ary = []
    opt_ary.each do |opt|
      next  unless opt
      if opt.class == Hash
        opt.each do |key, val|
          next  unless val
          if key.class != Symbol
            key = key.to_s.sub(/^\-+/, "").gsub(/\-/, "_").to_sym
          end
          if val.class != Symbol
            val = val.to_s.sub(/^\-+/, "").gsub(/\-/, "_").to_sym
          end
          hs[key] = val
        end
      else  # Hashł͂Ȃ
        opt = opt.to_s.sub(/^\-+/, "").gsub(/\-/, "_").to_sym
        ary << opt
      end
    end
    hs[:from] = input_type  unless hs[:from]
    hs[:to] = output_type  unless hs[:to]
    ary << hs
    output_str = PandocRuby.new(str, *ary).convert
    return output_str
  end

  def pandoc_opt(opt_ary)
    hs = {}
    var = []
    ary = []
    opt_ary.each do |opt|
      next  unless opt
      if opt.class == Hash
        opt.each do |key, val|
          next  unless val
          if key.class != Symbol
            key = key.to_s.sub(/^\-+/, "").gsub(/\-/, "_").to_sym
          end
          if key == :V or key == :variable
            var << val
            next
          end
          if val.class != Symbol
            val = val.to_s.sub(/^\-+/, "").gsub(/\-/, "_").to_sym
          end
          hs[key] = val
        end
      else  # Hashł͂Ȃ
        opt = opt.to_s.sub(/^\-+/, "").gsub(/\-/, "_").to_sym
        ary << opt
      end
    end
    return [hs, var, ary]
  end

  def pandoc_docx(str, *opt_ary)
    unless defined?(PandocRuby)
      STDERR.puts "'pandoc-ruby' has not loaded."
      return nil
    end
    hs, var, ary = pandoc_opt(opt_ary)
    hs[:to] = :docx
    ary << hs
    var.each {|x|  ary.push({:V=>x})}
    return pandoc_convert(str, *ary)
  end

  def pandoc_html(str, *opt_ary)
    unless defined?(PandocRuby)
      STDERR.puts "'pandoc-ruby' has not loaded."
      return nil
    end
    hs, var, ary = pandoc_opt(opt_ary)
    hs[:to] = :html
    ary << hs
    var.each {|x|  ary.push({:V=>x})}
    return pandoc_convert(str, *ary)
  end

  def pandoc_tex(str, *opt_ary)
    unless defined?(PandocRuby)
      STDERR.puts "'pandoc-ruby' has not loaded."
      return nil
    end
    hs, var, ary = pandoc_opt(opt_ary)
    hs[:to] = :latex
    ary << hs
    dclass = false
    var.each do |x|
      dclass = true  if x.to_s =~ /^documentclass/i
    end
    var << "documentclass:jarticle"  unless dclass
    var.each {|x|  ary.push({:V=>x})}
    tex_str = pandoc_convert(str, *ary)
    return tex_str
  end

  # xxx, yyy2, zzz_a, zzz["key"] Ȃǂ̂Rd˃p^[ERBpɕϊ
      # u<% cc %>v̕ɂ͕ϊȂ
  def erb_three(str)
    str = str.gsub(/\<\%/, "\x01\x7f")
    str = str.gsub(/\%\>/, "\x7f\x02")
    ary = []
    while str =~ /\x01\x7f([^\x7f]+)\x7f\x02/
      ptn = $1
      ary << ptn
      rstr = sprintf("\x19\x7f%d\x7f\x1a", ary.size-1)
      str = str.sub(/\x01\x7f[^\x7f]+\x7f\x02/, rstr)
    end
    reg = /([a-z])\1\1(\d+|_[A-Za-z0-9]+|\[.+?\])*/  # rrrȂǂ̌op^[
    str = str.gsub(reg) do |ptn|
      ptn_str = ptn.inspect
      ptn = "<%= (!defined?(#{ptn}) || #{ptn} == nil) ? #{ptn_str} : #{ptn} %>"
    end
    while str =~ /\x19\x7f(\d+)\x7f\x1a/
      n = $1.to_i
      ptn = ary[n]
      str = str.sub(/\x19\x7f\d+\x7f\x1a/, "<%#{ptn}%>")
    end
    str = str.gsub(/\x01\x7f/, "<%")
    str = str.gsub(/\x7f\x02/, "%>")
    return str
  end

  module_function(:kramdown_html, :kramdown_tex, :kramdown_contents,
      :pandoc_convert, :pandoc_opt, :pandoc_docx,
      :pandoc_html, :pandoc_tex, :erb_three)
end

Rrx::temp_delete_all()
